# Used in the following published models:
# nyu-visionx/cambrian-phi3-3b
# nyu-visionx/cambrian-8b
# nyu-visionx/cambrian-13b
# nyu-visionx/cambrian-34b

import torch
import torch.nn.functional as F
from open_clip import create_model_from_pretrained
from timm.models.convnext import ConvNeXt

from oumi.models.experimental.cambrian.model.multimodal_encoder.base_encoder import (
    BaseVisionTower,
    ProcessorWrapper,
)


def extract_res_interp(model_name):
    valid_model_prefixes = {
        "clip-convnext-L": "hf-hub:laion/CLIP-convnext_large_d_320.laion2B-s29B-b131K-ft-soup",
        "clip-convnext-XXL": "hf-hub:laion/CLIP-convnext_xxlarge-laion2B-s34B-b82K-augreg-soup",
    }

    res = None
    interp = None

    for prefix in valid_model_prefixes:
        if model_name.startswith(prefix):
            base_model_name = valid_model_prefixes[prefix]
            break
    else:
        raise ValueError(f"Unknown vision tower: {model_name}")

    parts = model_name.split("-")
    for part in parts:
        if part.startswith("res"):
            res = int(part[3:])
        elif part.startswith("interp"):
            interp = int(part[6:])

    return base_model_name, res, interp


class CLIPConvNextTower(BaseVisionTower):
    def __init__(self, vision_tower, args, delay_load=False):
        """Initialize the CLIPConvNextTower.

        Args:
            vision_tower (str): The name of the vision tower model in the format "clip-convnext-resXXX-interpYYY".
            args (argparse.Namespace): The arguments parsed from the command line.
            delay_load (bool, optional): Whether to delay loading the model. Defaults to False.
        """
        super().__init__(vision_tower, args, delay_load)

        self.is_multi_stage = "multi-stage" in vision_tower
        base_model_name, res, interp = extract_res_interp(vision_tower)
        self.vision_tower_name = base_model_name
        self._image_size = res if res is not None else 1024
        self._interp_size = interp  # default 256
        self._reduction = 32

        self.select_layer = args.mm_vision_select_layer
        self.select_feature = getattr(args, "mm_vision_select_feature", "patch")
        self.unfreeze_mm_vision_tower = getattr(args, "unfreeze_mm_vision_tower", False)
        self.is_loaded = False

        if not delay_load:
            self.load_model()
        elif self.unfreeze_mm_vision_tower:
            self.load_model()
        else:
            assert (
                "clip-convnext-L" in vision_tower or "clip-convnext-XXL" in vision_tower
            )
            if "clip-convnext-L" in vision_tower:
                if "multi-stage" in vision_tower:
                    self._hidden_size = sum([192, 384, 768, 1536])
                else:
                    self._hidden_size = 1536
            else:
                if "multi-stage" in vision_tower:
                    self._hidden_size = sum([384, 768, 1536, 3072])
                else:
                    self._hidden_size = 3072

    def load_model(self, device_map=None):
        """Load the CLIP-ConvNext model."""
        assert "clip-convnext" in self.vision_tower_name.lower()
        self.vision_model = "convnext"
        clip_model, processor = create_model_from_pretrained(self.vision_tower_name)
        processor.transforms[0].size = self._image_size
        processor.transforms[1].size = (self._image_size, self._image_size)
        self.image_processor = ProcessorWrapper(
            processor, height=self._image_size, width=self._image_size
        )
        self.vision_tower: ConvNeXt = clip_model.visual.trunk
        self.vision_tower.output_tokens = True
        feature_info = self.vision_tower.feature_info
        if self.is_multi_stage:
            self._hidden_size = sum([stage["num_chs"] for stage in feature_info])
        else:
            self._hidden_size = feature_info[-1]["num_chs"]
        self.vision_tower.requires_grad_(self.unfreeze_mm_vision_tower)
        self.is_loaded = True

    def interpolate(self, image_forward_outs):
        """Interpolate the image features to the desired number of patches.

        Args:
            image_forward_outs (torch.Tensor): The output features from the vision tower.

        Returns:
            torch.Tensor: The interpolated image features.
        """
        if self._interp_size is None:
            return image_forward_outs

        image_features = F.interpolate(
            image_forward_outs.float(),
            size=(self.num_patches_per_side, self.num_patches_per_side),
            mode="bilinear",
            align_corners=False,
        ).to(dtype=image_forward_outs.dtype)
        image_features = image_features.flatten(2, 3).permute(0, 2, 1).contiguous()
        return image_features

    def _forward(self, images):
        """Perform the forward pass of the CLIPConvNextTower.

        Args:
            images (torch.Tensor): The input images.

        Returns:
            torch.Tensor: The output features from the vision tower after interpolation.
        """
        with torch.set_grad_enabled(self.unfreeze_mm_vision_tower):
            image_features_stages = []
            x = self.vision_tower.stem(images.to(device=self.device, dtype=self.dtype))
            for stage in self.vision_tower.stages:
                x = stage(x)
                image_features_stages.append(x)
            if not self.is_multi_stage:
                image_features_stages = image_features_stages[-1:]
            image_features_stages_rescaled = []
            for image_features_single_stage in image_features_stages:
                image_features_single_stage_rescaled = self.interpolate(
                    image_features_single_stage
                )
                image_features_stages_rescaled.append(
                    image_features_single_stage_rescaled
                )
            image_features = torch.cat(image_features_stages_rescaled, -1)
            return image_features

    @property
    def image_size(self):
        return self._image_size

    @property
    def num_patches_per_side(self):
        """Get the number of patches per side.

        Returns:
            int: The number of patches per side.
        """
        if self._interp_size is None:
            return self._image_size // self._reduction
        else:
            return int(self._interp_size**0.5)

    @property
    def num_patches(self):
        """Get the total number of patches.

        Default: 256

        Returns:
            int: The total number of patches.
        """
        if self._interp_size is None:
            return (self._image_size // self._reduction) ** 2
        else:
            return self._interp_size



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import requests
import yaml
from typing_extensions import Self

from oumi.cli import cli_utils
from oumi.cli.alias import AliasType, try_get_config_name_for_alias
from oumi.core.configs import BaseConfig
from oumi.core.configs.inference_config import InferenceConfig
from oumi.core.configs.params.judge_params import JudgeParams

JUDGE_CONFIG_REPO_PATH_TEMPLATE = "oumi://configs/projects/judges/{path}.yaml"


@dataclass
class JudgeConfig(BaseConfig):
    """Consolidated configuration for the Judge.

    This class combines the judge parameters (JudgeParams) and inference
    configuration (InferenceConfig) into a single configuration object.

    Example:
        >>> judge_config = JudgeConfig( # doctest: +SKIP
        ...     judge_params=JudgeParams(
        ...         prompt_template="Is this helpful? {question}, {answer}",
        ...         response_format=JudgeResponseFormat.XML,
        ...         judgment_type=JudgeOutputType.BOOL,
        ...         include_explanation=False
        ...     ),
        ...     inference_config=InferenceConfig(
        ...         model=ModelParams(model_name="gpt-4.1"),
        ...         generation=GenerationParams(max_tokens=100),
        ...         engine=InferenceEngineType.OPENAI
        ...     )
        ... )
    """

    judge_params: JudgeParams
    """Parameters for the judge prompt and response format."""

    inference_config: InferenceConfig
    """Configuration for the inference engine and generation parameters."""

    @classmethod
    def from_path(cls, path: str, extra_args: Optional[list[str]] = None) -> Self:
        """Resolve the JudgeConfig from a local or repo path."""

        def _resolve_path(unresolved_path: str) -> Optional[str]:
            try:
                # Attempt to resolve the path using CLI utilities.
                # This will handle both local paths and repo (oumi://) paths.
                resolved_path = str(
                    cli_utils.resolve_and_fetch_config(
                        unresolved_path,
                    )
                )
            except (
                requests.exceptions.RequestException,  # Network/HTTP issues
                yaml.YAMLError,  # YAML parsing errors
                OSError,  # File system operations (includes IOError)
            ):
                # If resolution fails, mask the error and return None.
                return None

            # If resolution succeeds, check if the resolved path exists indeed.
            return resolved_path if Path(resolved_path).exists() else None

        if extra_args is None:
            extra_args = []

        # If `path` is an alias, resolve it to the corresponding oumi:// path.
        path = try_get_config_name_for_alias(path, AliasType.JUDGE)

        # If `path` is a local or repo path, load JudgeConfig obj from that path.
        # Repo example: path = "oumi://configs/projects/judges/doc_qa/relevance.yaml"
        # Local example: path= "./local_path/relevance.yaml"
        resolved_path = _resolve_path(path)

        # If `path` is a built-in judge name, construct the path from the default
        # repo location, and then load the corresponding JudgeConfig.
        # Example:
        # "doc_qa/relevance" => "oumi://configs/projects/judges/doc_qa/relevance.yaml"
        if not resolved_path:
            resolved_path = _resolve_path(
                JUDGE_CONFIG_REPO_PATH_TEMPLATE.format(path=path)
            )

        if resolved_path:
            try:
                return cls.from_yaml_and_arg_list(resolved_path, extra_args)
            except Exception as e:
                raise ValueError(
                    f"Failed to parse {resolved_path} as JudgeConfig. "
                    f"Please ensure the YAML file contains both 'judge_params' and "
                    f"'inference_config' sections with valid fields. "
                    f"Original error: {e}"
                ) from e
        else:
            raise ValueError(
                f"Could not resolve JudgeConfig from path: {path}. "
                "Please provide a valid local or GitHub repo path."
            )



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
from typing_extensions import override

from oumi.core.datasets.base_grpo_dataset import BaseExperimentalGrpoDataset
from oumi.core.registry import register_dataset
from oumi.core.types.conversation import Conversation

_SYSTEM_PROMPT = (
    "Your final answer should be formatted as "
    r'```json YOUR_ANSWER```. For example, if the answer is {"a": 1}, '
    r'you should output ```json {"a": 1}}```.'
)


@register_dataset("oumi-ai/berrybench-v0.1.1")
class BerryBenchGrpoDataset(BaseExperimentalGrpoDataset):
    r"""Dataset class for the `oumi-ai/berrybench-v0.1.1` dataset.

    A sample from the dataset::

        {
            "messages": [
                {
                    "content": "Return a JSON object showing the frequency of each character in the word '黒い'. Only include characters that appear in the word.",
                    "role": "user",
                }
            ],
            "metadata": {
                "character_count": 2,
                "difficulty": 3,
                "expected_response": '{"\\u9ed2": 1, "\\u3044": 1}',
                "language": "japanese",
                "word": "黒い",
            },
        }
    """  # noqa: E501

    default_dataset = "oumi-ai/berrybench-v0.1.1"

    @override
    def transform(self, sample: pd.Series) -> dict:
        """Transform the sample into Python `dict`."""
        # Add system prompt before user prompt.
        system_message = {"content": _SYSTEM_PROMPT, "role": "system"}
        messages = [system_message, sample["messages"][0]]
        return {
            "prompt": messages,
            "metadata": sample["metadata"],
        }

    @override
    def transform_conversation(self, sample: pd.Series) -> Conversation:
        """Converts the input sample to a Conversation.

        Args:
            sample (dict): The input example.

        Returns:
            Conversation: The resulting conversation.

        """
        # Example is already in conversation format and only needs light processing.
        sample_dict = sample.to_dict()
        # Add system prompt before user prompt.
        system_message = {"content": _SYSTEM_PROMPT, "role": "system"}
        messages = [system_message, sample["messages"][0]]
        sample_dict["messages"] = messages
        return Conversation.from_dict(sample_dict)



from unittest.mock import Mock, call, patch

import pytest

from oumi.core.configs import JobConfig, JobResources, StorageMount
from oumi.core.launcher import JobState, JobStatus
from oumi.launcher.clients.local_client import LocalClient
from oumi.launcher.clusters.local_cluster import LocalCluster


#
# Fixtures
#
@pytest.fixture
def mock_local_client():
    yield Mock(spec=LocalClient)


def _get_default_job(cloud: str) -> JobConfig:
    resources = JobResources(
        cloud=cloud,
        region="us-central1",
        zone=None,
        accelerators="A100-80GB",
        cpus="4",
        memory="64",
        instance_type=None,
        use_spot=True,
        disk_size=512,
        disk_tier="low",
    )
    return JobConfig(
        name="myjob",
        user="user",
        working_dir="./",
        num_nodes=2,
        resources=resources,
        envs={"var1": "val1"},
        file_mounts={
            "~/home/remote/path.bar": "~/local/path.bar",
            "~/home/remote/path2.txt": "~/local/path2.txt",
        },
        storage_mounts={
            "~/home/remote/path/gcs/": StorageMount(
                source="gs://mybucket/", store="gcs"
            )
        },
        setup=(
            "#PBS -o some/log \n#PBE -l wow\n#PBS -e run/log\n"
            "pip install -r requirements.txt"
        ),
        run="./hello_world.sh",
    )


#
# Tests
#
def test_local_cluster_name(mock_local_client):
    cluster = LocalCluster("local name", mock_local_client)
    assert cluster.name() == "local name"

    cluster = LocalCluster("", mock_local_client)
    assert cluster.name() == ""


def test_local_cluster_get_job_valid_id(mock_local_client):
    cluster = LocalCluster("name", mock_local_client)
    mock_local_client.list_jobs.return_value = [
        JobStatus(
            id="myjob",
            name="some name",
            status="running",
            metadata="",
            cluster="",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="job2",
            name="some",
            status="running",
            metadata="",
            cluster="",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="final job",
            name="name3",
            status="running",
            metadata="",
            cluster="",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    job = cluster.get_job("myjob")
    mock_local_client.list_jobs.assert_called_once()
    assert job == JobStatus(
        id="myjob",
        name="some name",
        status="running",
        metadata="",
        cluster="name",
        done=False,
        state=JobState.PENDING,
    )


def test_local_cluster_get_job_invalid_id_empty(mock_local_client):
    cluster = LocalCluster("name", mock_local_client)
    mock_local_client.list_jobs.return_value = []
    job = cluster.get_job("myjob")
    mock_local_client.list_jobs.assert_called_once()
    assert job is None


def test_local_cluster_get_job_invalid_id_nonempty(mock_local_client):
    cluster = LocalCluster("name", mock_local_client)
    mock_local_client.list_jobs.return_value = [
        JobStatus(
            id="myjob",
            name="some name",
            status="running",
            metadata="",
            cluster="mycluster",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="job2",
            name="some",
            status="running",
            metadata="",
            cluster="mycluster",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="final job",
            name="name3",
            status="running",
            metadata="",
            cluster="mycluster",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    job = cluster.get_job("wrong job")
    mock_local_client.list_jobs.assert_called_once()
    assert job is None


def test_local_cluster_get_jobs_nonempty(mock_local_client):
    cluster = LocalCluster("name", mock_local_client)
    mock_local_client.list_jobs.return_value = [
        JobStatus(
            id="myjob",
            name="some name",
            status="running",
            metadata="",
            cluster="mycluster",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="job2",
            name="some",
            status="running",
            metadata="",
            cluster="mycluster",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="final job",
            name="name3",
            status="running",
            metadata="",
            cluster="mycluster",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    jobs = cluster.get_jobs()
    mock_local_client.list_jobs.assert_called_once()
    expected_jobs = [
        JobStatus(
            id="myjob",
            name="some name",
            status="running",
            metadata="",
            cluster="name",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="job2",
            name="some",
            status="running",
            metadata="",
            cluster="name",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="final job",
            name="name3",
            status="running",
            metadata="",
            cluster="name",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    assert jobs == expected_jobs


def test_local_cluster_get_jobs_empty(mock_local_client):
    cluster = LocalCluster("name", mock_local_client)
    mock_local_client.list_jobs.return_value = []
    jobs = cluster.get_jobs()
    mock_local_client.list_jobs.assert_called_once()
    expected_jobs = []
    assert jobs == expected_jobs


def test_local_cluster_cancel_job(mock_local_client):
    cluster = LocalCluster("name", mock_local_client)
    mock_local_client.list_jobs.return_value = [
        JobStatus(
            id="myjob",
            name="some name",
            status="running",
            metadata="",
            cluster="debug.name",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="job2",
            name="some",
            status="running",
            metadata="",
            cluster="debug.name",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="final job",
            name="name3",
            status="running",
            metadata="",
            cluster="debug.name",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    job_status = cluster.cancel_job("job2")
    expected_status = JobStatus(
        id="job2",
        name="some",
        status="running",
        metadata="",
        cluster="name",
        done=False,
        state=JobState.PENDING,
    )
    mock_local_client.cancel.assert_called_once_with(
        "job2",
    )
    assert job_status == expected_status


def test_local_cluster_cancel_job_fails(mock_local_client):
    cluster = LocalCluster("name", mock_local_client)
    mock_local_client.list_jobs.return_value = [
        JobStatus(
            id="job2",
            name="some",
            status="running",
            metadata="",
            cluster="debug.name",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    with pytest.raises(RuntimeError, match="Job myjobid not found."):
        _ = cluster.cancel_job("myjobid")


def test_local_cluster_run_job(mock_local_client):
    cluster = LocalCluster("name", mock_local_client)
    mock_local_client.submit_job.return_value = JobStatus(
        id="1234",
        name="some name",
        status="queued",
        metadata="",
        cluster="mycluster",
        done=False,
        state=JobState.PENDING,
    )
    expected_status = JobStatus(
        id="1234",
        name="some name",
        status="queued",
        metadata="",
        cluster="name",
        done=False,
        state=JobState.PENDING,
    )
    expected_job = _get_default_job("local")
    job_status = cluster.run_job(expected_job)
    mock_local_client.submit_job.assert_called_once_with(
        expected_job,
    )
    assert job_status == expected_status


def test_local_cluster_run_job_no_name(mock_local_client):
    cluster = LocalCluster("name", mock_local_client)
    mock_local_client.submit_job.return_value = JobStatus(
        id="1234",
        name="some name",
        status="queued",
        metadata="",
        cluster="mycluster",
        done=False,
        state=JobState.PENDING,
    )
    expected_status = JobStatus(
        id="1234",
        name="some name",
        status="queued",
        metadata="",
        cluster="name",
        done=False,
        state=JobState.PENDING,
    )
    job = _get_default_job("local")
    job.name = None
    with patch("oumi.launcher.clusters.local_cluster.uuid") as mock_uuid:
        mock_hex = Mock()
        mock_hex.hex = "1-2-3"
        mock_uuid.uuid1.return_value = mock_hex
        job_status = cluster.run_job(job)
    expected_job = _get_default_job("local")
    expected_job.name = "1-2-3"
    mock_local_client.submit_job.assert_called_once_with(
        expected_job,
    )
    assert job_status == expected_status


def test_local_cluster_down(mock_local_client):
    cluster = LocalCluster("name", mock_local_client)
    mock_local_client.list_jobs.side_effect = [
        [
            JobStatus(
                id="myjob",
                name="some name",
                status="running",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="job2",
                name="some",
                status="running",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="final job",
                name="name3",
                status="running",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
        ],
        [
            JobStatus(
                id="myjob",
                name="some name",
                status="CANCELED",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="job2",
                name="some",
                status="running",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="final job",
                name="name3",
                status="running",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
        ],
        [
            JobStatus(
                id="myjob",
                name="some name",
                status="CANCELED",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="job2",
                name="some",
                status="CANCELED",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="final job",
                name="name3",
                status="running",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
        ],
        [
            JobStatus(
                id="myjob",
                name="some name",
                status="CANCELED",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="job2",
                name="some",
                status="CANCELED",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="final job",
                name="name3",
                status="CANCELED",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
        ],
    ]
    cluster.down()
    mock_local_client.cancel.assert_has_calls(
        [call("myjob"), call("job2"), call("final job")]
    )
    mock_local_client.list_jobs.assert_has_calls([call(), call(), call(), call()])
    # Nothing to assert, this method is a no-op.


def test_local_cluster_stop(mock_local_client):
    cluster = LocalCluster("name", mock_local_client)
    mock_local_client.list_jobs.side_effect = [
        [
            JobStatus(
                id="myjob",
                name="some name",
                status="running",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="job2",
                name="some",
                status="running",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="final job",
                name="name3",
                status="running",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
        ],
        [
            JobStatus(
                id="myjob",
                name="some name",
                status="CANCELED",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="job2",
                name="some",
                status="running",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="final job",
                name="name3",
                status="running",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
        ],
        [
            JobStatus(
                id="myjob",
                name="some name",
                status="CANCELED",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="job2",
                name="some",
                status="CANCELED",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="final job",
                name="name3",
                status="running",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
        ],
        [
            JobStatus(
                id="myjob",
                name="some name",
                status="CANCELED",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="job2",
                name="some",
                status="CANCELED",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
            JobStatus(
                id="final job",
                name="name3",
                status="CANCELED",
                metadata="",
                cluster="",
                done=False,
                state=JobState.PENDING,
            ),
        ],
    ]
    cluster.stop()
    mock_local_client.cancel.assert_has_calls(
        [call("myjob"), call("job2"), call("final job")]
    )
    mock_local_client.list_jobs.assert_has_calls([call(), call(), call(), call()])
    # Nothing to assert, this method is a no-op.



import tempfile
import time
from datetime import datetime, timezone
from io import TextIOWrapper
from subprocess import PIPE
from unittest.mock import Mock, call, patch

import pytest

from oumi.core.configs import JobConfig, JobResources, StorageMount
from oumi.core.launcher import JobState, JobStatus
from oumi.launcher.clients.local_client import LocalClient


#
# Fixtures
#
@pytest.fixture
def mock_time():
    with patch("oumi.launcher.clients.local_client.time") as time_mock:
        yield time_mock


@pytest.fixture
def mock_thread():
    with patch("oumi.launcher.clients.local_client.Thread") as thread_mock:
        yield thread_mock


@pytest.fixture
def mock_popen():
    with patch("oumi.launcher.clients.local_client.Popen") as popen_mock:
        yield popen_mock


@pytest.fixture
def mock_os():
    with patch("oumi.launcher.clients.local_client.os") as os_mock:
        yield os_mock


@pytest.fixture
def mock_datetime():
    with patch("oumi.launcher.clients.local_client.datetime") as datetime_mock:
        yield datetime_mock


class OpenEquivalent(TextIOWrapper):
    def __init__(self, file):
        self.file = file

    def __eq__(self, other):
        return (
            self.file.mode == other.mode
            and self.file.name == other.name
            and self.file.encoding == other.encoding
        )


def _get_default_job() -> JobConfig:
    resources = JobResources(
        cloud="local",
        region="us-central1",
        zone=None,
        accelerators="A100-80",
        cpus="4",
        memory="64",
        instance_type=None,
        use_spot=True,
        disk_size=512,
        disk_tier="low",
    )
    return JobConfig(
        name="myjob",
        user="user",
        working_dir="./",
        num_nodes=2,
        resources=resources,
        envs={"var1": "val1"},
        file_mounts={},
        storage_mounts={
            "~/home/remote/path/gcs/": StorageMount(
                source="gs://mybucket/", store="gcs"
            )
        },
        setup="pip install -r requirements.txt",
        run="./hello_world.sh",
    )


#
# Tests
#
def test_local_client_init(mock_thread):
    client = LocalClient()
    mock_thread.assert_called_once_with(target=client._worker_loop, daemon=True)


def test_local_client_submit_job(mock_thread):
    client = LocalClient()
    job = _get_default_job()
    job_status = client.submit_job(job)
    expected_status = JobStatus(
        id="0",
        name=str(job.name),
        cluster="",
        status="QUEUED",
        metadata="",
        done=False,
        state=JobState.PENDING,
    )
    assert job_status == expected_status
    assert client.list_jobs() == [expected_status]


def test_local_client_submit_job_execution(
    mock_time, mock_popen, mock_os, mock_datetime
):
    client = LocalClient()
    mock_os.environ.copy.return_value = {"preset": "value"}
    mock_datetime.fromtimestamp.return_value = datetime.fromtimestamp(10)
    mock_time.time.return_value = 10
    mock_process = Mock()
    mock_popen.return_value = mock_process
    mock_process.wait.return_value = 0
    job = _get_default_job()
    job_status = client.submit_job(job)
    time.sleep(1)
    expected_cmds = """cd ./
pip install -r requirements.txt
./hello_world.sh"""
    mock_popen.assert_called_once_with(
        expected_cmds,
        shell=True,
        env={"preset": "value", "var1": "val1"},
        stdout=PIPE,
        stderr=PIPE,
    )
    expected_status = JobStatus(
        id="0",
        name=str(job.name),
        cluster="",
        status="COMPLETED",
        metadata=f"Job finished at {datetime.fromtimestamp(10).isoformat()} .",
        done=True,
        state=JobState.SUCCEEDED,
    )
    assert job_status == expected_status
    assert client.list_jobs() == [expected_status]


def test_local_client_submit_job_execution_with_logging(
    mock_time, mock_popen, mock_os, mock_datetime
):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        client = LocalClient()
        mock_os.environ.copy.return_value = {
            "preset": "value",
            "OUMI_LOGGING_DIR": output_temp_dir,
        }
        mock_datetime.fromtimestamp.return_value = datetime.fromtimestamp(10)
        mock_datetime.now.return_value = datetime.fromtimestamp(10.1234, timezone.utc)
        mock_time.time.return_value = 10
        mock_process = Mock()
        mock_popen.return_value = mock_process
        mock_process.wait.return_value = 0
        job = _get_default_job()
        job_status = client.submit_job(job)
        time.sleep(1)
        expected_cmds = """cd ./
pip install -r requirements.txt
./hello_world.sh"""
        stdout_handler = open(
            f"{output_temp_dir}/1970_01_01_00_00_10_123_0.stdout", "w"
        )
        std_err_handler = open(
            f"{output_temp_dir}/1970_01_01_00_00_10_123_0.stderr", "w"
        )
        mock_popen.assert_called_once_with(
            expected_cmds,
            shell=True,
            env={
                "preset": "value",
                "OUMI_LOGGING_DIR": output_temp_dir,
                "var1": "val1",
            },
            stdout=OpenEquivalent(stdout_handler),
            stderr=OpenEquivalent(std_err_handler),
        )
        expected_status = JobStatus(
            id="0",
            name=str(job.name),
            cluster="",
            status="COMPLETED",
            metadata=f"Job finished at {datetime.fromtimestamp(10).isoformat()} ."
            f" Logs available at: {output_temp_dir}/1970_01_01_00_00_10_123_0.stdout",
            done=True,
            state=JobState.SUCCEEDED,
        )
        assert job_status == expected_status
        assert client.list_jobs() == [expected_status]


def test_local_client_submit_job_execution_multiple(
    mock_time, mock_popen, mock_os, mock_datetime
):
    client = LocalClient()
    mock_datetime.fromtimestamp.return_value = datetime.fromtimestamp(10)
    mock_os.environ.copy.side_effect = [{"preset": "value"}, {"preset": "value"}]
    mock_time.time.return_value = 10
    mock_process = Mock()
    mock_process.wait.return_value = 0
    mock_process2 = Mock()
    mock_process2.wait.return_value = 1
    mock_process2.stderr = "hidden message" + "a" * 1024
    mock_popen.side_effect = [mock_process, mock_process2]
    job = _get_default_job()
    second_job = _get_default_job()
    second_job.envs = {}
    second_job.setup = None
    second_job.run = "echo 'hello'"
    second_job.working_dir = "~/second"
    job_status = client.submit_job(job)
    second_job_status = client.submit_job(second_job)
    time.sleep(1)
    expected_cmds = """cd ./
pip install -r requirements.txt
./hello_world.sh"""
    second_expected_cmds = """cd ~/second

echo 'hello'"""
    mock_popen.assert_has_calls(
        [
            call(
                expected_cmds,
                shell=True,
                env={"preset": "value", "var1": "val1"},
                stdout=PIPE,
                stderr=PIPE,
            ),
            call(
                second_expected_cmds,
                shell=True,
                env={"preset": "value"},
                stdout=PIPE,
                stderr=PIPE,
            ),
        ]
    )
    expected_status = JobStatus(
        id="0",
        name=str(job.name),
        cluster="",
        status="COMPLETED",
        metadata=f"Job finished at {datetime.fromtimestamp(10).isoformat()} .",
        done=True,
        state=JobState.SUCCEEDED,
    )
    second_expected_status = JobStatus(
        id="1",
        name=str(second_job.name),
        cluster="",
        status="FAILED",
        metadata="a" * 1024,
        done=True,
        state=JobState.FAILED,
    )
    assert job_status == expected_status
    assert second_job_status == second_expected_status
    assert client._running_process is None
    assert client.list_jobs() == [expected_status, second_expected_status]


def test_local_client_list_jobs(mock_thread):
    client = LocalClient()
    job = _get_default_job()
    job_status = client.submit_job(job)
    assert client.list_jobs() == [job_status]


def test_local_client_list_jobs_multiple(mock_thread):
    client = LocalClient()
    job = _get_default_job()
    job2 = _get_default_job()
    job2.name = "secondjob"
    job3 = _get_default_job()
    job3.name = "thirdjob"
    job_status = client.submit_job(job)
    job_status2 = client.submit_job(job2)
    job_status3 = client.submit_job(job3)
    assert client.list_jobs() == [job_status, job_status2, job_status3]
    assert job_status.id == "0"
    assert job_status2.id == "1"
    assert job_status3.id == "2"
    assert job_status.name == "myjob"
    assert job_status2.name == "secondjob"
    assert job_status3.name == "thirdjob"


def test_local_client_get_job(mock_thread):
    client = LocalClient()
    job = _get_default_job()
    job2 = _get_default_job()
    job2.name = "secondjob"
    job3 = _get_default_job()
    job3.name = "thirdjob"
    job_status = client.submit_job(job)
    job_status2 = client.submit_job(job2)
    job_status3 = client.submit_job(job3)
    assert client.get_job("0") == job_status
    assert client.get_job("1") == job_status2
    assert client.get_job("2") == job_status3
    client.cancel("1")
    updated_status = client.get_job("1")
    assert updated_status is not None
    assert updated_status.status == "CANCELED"
    assert client.get_job("3") is None


def test_local_client_cancel(mock_thread):
    client = LocalClient()
    job = _get_default_job()
    job2 = _get_default_job()
    job2.name = "secondjob"
    job_status = client.submit_job(job)
    job_status2 = client.submit_job(job2)
    assert job_status.status == "QUEUED"
    assert job_status2.status == "QUEUED"
    job_status2 = client.cancel("1")
    assert job_status.status == "QUEUED"
    assert job_status2 is not None
    assert job_status2.status == "CANCELED"
    job_status = client.cancel("0")
    assert job_status is not None
    assert job_status.status == "CANCELED"
    assert job_status2 is not None
    assert job_status2.status == "CANCELED"
    assert client.cancel("2") is None


def test_local_client_cancel_completed_jobs(mock_time, mock_popen, mock_os):
    client = LocalClient()
    mock_os.environ.copy.side_effect = [{"preset": "value"}, {"preset": "value"}]
    mock_time.time.return_value = 10
    mock_process = Mock()
    mock_process.wait.return_value = 0
    mock_process2 = Mock()
    mock_process2.wait.return_value = 1
    mock_process2.stderr = "hidden message" + "a" * 1024
    mock_popen.side_effect = [mock_process, mock_process2]
    job = _get_default_job()
    second_job = _get_default_job()
    second_job.envs = {}
    second_job.setup = None
    second_job.run = "echo 'hello'"
    second_job.working_dir = "~/second"
    job_status = client.submit_job(job)
    second_job_status = client.submit_job(second_job)
    time.sleep(1)
    job_status = client.get_job("0")
    second_job_status = client.get_job("1")
    assert job_status is not None
    assert second_job_status is not None
    assert job_status.status == "COMPLETED"
    assert second_job_status.status == "FAILED"

    job_status_after_cancel = client.cancel("0")
    job_status2_after_cancel = client.cancel("1")
    assert job_status_after_cancel is not None
    assert job_status2_after_cancel is not None
    assert job_status_after_cancel.status == "COMPLETED"
    assert job_status2_after_cancel.status == "FAILED"



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from pathlib import Path
from typing import Optional, cast

from tqdm.auto import tqdm
from typing_extensions import override

from oumi.core.configs import GenerationParams, InferenceConfig, ModelParams
from oumi.core.inference import BaseInferenceEngine
from oumi.core.types.conversation import Conversation, Message, Role
from oumi.utils.logging import logger

try:
    from llama_cpp import Llama  # pyright: ignore[reportMissingImports]
except ModuleNotFoundError:
    Llama = None


class LlamaCppInferenceEngine(BaseInferenceEngine):
    """Engine for running llama.cpp inference locally.

    This class provides an interface for running inference using the llama.cpp library
    on local hardware. It allows for efficient execution of large language models
    with quantization, kv-caching, prefix filling, ...

    Note:
        This engine requires the llama-cpp-python package to be installed.
        If not installed, it will raise a RuntimeError.

    Example:
        >>> from oumi.core.configs import ModelParams
        >>> from oumi.inference import LlamaCppInferenceEngine
        >>> model_params = ModelParams(
        ...     model_name="path/to/model.gguf",
        ...     model_kwargs={
        ...         "n_gpu_layers": -1,
        ...         "n_threads": 8,
        ...         "flash_attn": True
        ...     }
        ... )
        >>> engine = LlamaCppInferenceEngine(model_params) # doctest: +SKIP
        >>> # Use the engine for inference
    """

    def __init__(
        self,
        model_params: ModelParams,
        *,
        generation_params: Optional[GenerationParams] = None,
    ):
        """Initializes the LlamaCppInferenceEngine.

        This method sets up the engine for running inference using llama.cpp.
        It loads the specified model and configures the inference parameters.

        Documentation: https://llama-cpp-python.readthedocs.io/en/latest/api-reference/#llama_cpp.Llama.create_completion

        Args:
            model_params (ModelParams): Parameters for the model, including the model
                name, maximum length, and any additional keyword arguments for model
                initialization.
            generation_params (GenerationParams): Parameters for generation.

        Raises:
            RuntimeError: If the llama-cpp-python package is not installed.
            ValueError: If the specified model file is not found.

        Note:
            This method automatically sets some default values for model initialization:
            - verbose: False (reduces log output for bulk inference)
            - n_gpu_layers: -1 (uses GPU acceleration for all layers if available)
            - n_threads: 4
            - filename: "*q8_0.gguf" (applies Q8 quantization by default)
            - flash_attn: True
            - use_mmap: True (loads model parts as needed)
            - use_mlock: True (locks the model pages in physical RAM)
            These defaults can be overridden by specifying them in
            `model_params.model_kwargs`.
        """
        super().__init__(model_params=model_params, generation_params=generation_params)

        if not Llama:
            raise RuntimeError(
                "llama-cpp-python is not installed. "
                "Please install it with 'pip install llama-cpp-python'."
            )

        # `model_max_length` is required by llama-cpp, but optional in our config
        # Use a default value if not set.
        if model_params.model_max_length is None:
            model_max_length = 4096
            logger.warning(
                "model_max_length is not set. "
                f"Using default value of {model_max_length}."
            )
        else:
            model_max_length = model_params.model_max_length

        # Set some reasonable defaults. These will be overriden by the user if set in
        # the config.
        kwargs = {
            # llama-cpp logs a lot of useful information,
            # but it's too verbose by default for bulk inference.
            "verbose": False,
            # Put all layers on GPU / MPS if available. Otherwise, will use CPU.
            "n_gpu_layers": -1,
            # Increase the default number of threads.
            # Too many can cause deadlocks
            "n_threads": 4,
            # Use Q8 quantization by default.
            "filename": "*8_0.gguf",
            "flash_attn": True,
            # Memory safety defaults
            "use_mmap": True,
            "use_mlock": True,
        }

        model_kwargs = model_params.model_kwargs.copy()
        kwargs.update(model_kwargs)

        # Load model
        if Path(model_params.model_name).exists():
            logger.info(f"Loading model from disk: {model_params.model_name}.")
            kwargs.pop("filename", None)  # only needed if downloading from hub
            self._llm = Llama(
                model_path=model_params.model_name, n_ctx=model_max_length, **kwargs
            )
        else:
            logger.info(
                f"Loading model from Huggingface Hub: {model_params.model_name}."
            )
            self._llm = Llama.from_pretrained(
                repo_id=model_params.model_name, n_ctx=model_max_length, **kwargs
            )

    def _convert_conversation_to_llama_input(
        self, conversation: Conversation
    ) -> list[dict[str, str]]:
        """Converts a conversation to a list of llama.cpp input messages."""
        # FIXME Handle multimodal e.g., raise an error.
        role_mapping = {
            Role.SYSTEM: "system",
            Role.USER: "user",
            Role.ASSISTANT: "assistant",
        }
        return [
            {
                "content": message.compute_flattened_text_content(),
                "role": role_mapping.get(message.role, "assistant"),
            }
            for message in conversation.messages
        ]

    def _infer(
        self,
        input: list[Conversation],
        inference_config: Optional[InferenceConfig] = None,
    ) -> list[Conversation]:
        """Runs model inference on the provided input using llama.cpp.

        Args:
            input: A list of conversations to run inference on.
                Each conversation should contain at least one message.
            inference_config: Parameters for inference.

        Returns:
            List[Conversation]: A list of conversations with the model's responses
            appended. Each conversation in the output list corresponds to an input
            conversation, with an additional message from the assistant (model) added.
        """
        generation_params = (
            inference_config.generation
            if inference_config and inference_config.generation
            else self._generation_params
        )
        output_conversations = []

        # skip using a progress for single turns
        disable_tgdm = len(input) < 2

        for conversation in tqdm(input, disable=disable_tgdm):
            if not conversation.messages:
                logger.warning("Conversation must have at least one message.")
                # add the conversation to keep input and output the same length.
                output_conversations.append(conversation)
                continue

            llama_input = self._convert_conversation_to_llama_input(conversation)

            response = self._llm.create_chat_completion(
                messages=llama_input,  # type: ignore
                max_tokens=generation_params.max_new_tokens,
                temperature=generation_params.temperature,
                top_p=generation_params.top_p,
                frequency_penalty=generation_params.frequency_penalty,
                presence_penalty=generation_params.presence_penalty,
                stop=generation_params.stop_strings,
                logit_bias=generation_params.logit_bias,
                min_p=generation_params.min_p,
            )
            response = cast(dict, response)

            new_message = Message(
                content=response["choices"][0]["message"]["content"],
                role=Role.ASSISTANT,
            )

            messages = [
                *conversation.messages,
                new_message,
            ]
            new_conversation = Conversation(
                messages=messages,
                metadata=conversation.metadata,
                conversation_id=conversation.conversation_id,
            )
            output_conversations.append(new_conversation)
            self._save_conversation_to_scratch(
                new_conversation,
                inference_config.output_path if inference_config else None,
            )

        return output_conversations

    @override
    def get_supported_params(self) -> set[str]:
        """Returns a set of supported generation parameters for this engine."""
        return {
            "frequency_penalty",
            "logit_bias",
            "max_new_tokens",
            "min_p",
            "presence_penalty",
            "stop_strings",
            "temperature",
            "top_p",
        }

    def infer_online(
        self,
        input: list[Conversation],
        inference_config: Optional[InferenceConfig] = None,
    ) -> list[Conversation]:
        """Runs model inference online.

        Args:
            input: A list of conversations to run inference on.
            inference_config: Parameters for inference.

        Returns:
            List[Conversation]: Inference output.
        """
        warnings.warn(
            "infer_online() will be private in the future. Use infer() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        results = self._infer_online(input, inference_config)
        if inference_config and inference_config.output_path:
            self._save_conversations(results, inference_config.output_path)
        return results

    def infer_from_file(
        self,
        input_filepath: str,
        inference_config: Optional[InferenceConfig] = None,
    ) -> list[Conversation]:
        """Runs model inference on inputs in the provided file.

        This is a convenience method to prevent boilerplate from asserting the existence
        of input_filepath in the generation_params.

        Args:
            input_filepath: Path to the input file containing prompts for generation.
            inference_config: Parameters for inference.

        Returns:
            List[Conversation]: Inference output.
        """
        warnings.warn(
            "infer_from_file() will be private in the future. Use infer() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        input = self._read_conversations(input_filepath)
        results = self._infer(input, inference_config)
        if inference_config and inference_config.output_path:
            self._save_conversations(results, inference_config.output_path)
        return results

    @override
    def _infer_online(
        self,
        input: list[Conversation],
        inference_config: Optional[InferenceConfig] = None,
    ) -> list[Conversation]:
        """Runs model inference online.

        Args:
            input: A list of conversations to run inference on.
            inference_config: Parameters for inference.

        Returns:
            List[Conversation]: Inference output.
        """
        return self._infer(input, inference_config)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import torch
import transformers
from transformers import (
    get_cosine_schedule_with_warmup,
    get_cosine_with_hard_restarts_schedule_with_warmup,
    get_linear_schedule_with_warmup,
)

from oumi.core.configs import SchedulerType, TrainingParams
from oumi.utils.logging import logger


def build_lr_scheduler(
    optimizer: torch.optim.Optimizer,
    training_params: TrainingParams,
    num_training_steps: Optional[int] = None,
    current_epoch: int = 0,
) -> torch.optim.lr_scheduler.LRScheduler:
    """Builds a learning rate scheduler based on the provided training parameters.

    Args:
        optimizer: The optimizer for which to build the learning rate scheduler.
        training_params: The training parameters containing the scheduler args.
        num_training_steps: The total number of training steps
            (required for some schedulers).
        current_epoch (`int`, *optional*, defaults to 0):
            The index of the current epoch when resuming training.

    Optional Args (in `training_params.lr_scheduler_kwargs`):
        num_cycles (`float`, *optional*, defaults to 1.0 for cosine_with_restarts and
            0.5 for others): The number of cycles for the cosine family of schedulers.
        min_lr_rate (`float`, *optional*, defaults to 0.0): The minimum learning rate to
            decay to for cosine_with_min_lr.

    Returns:
        A learning rate scheduler.
    """
    scheduler_type = training_params.lr_scheduler_type.lower()
    scheduler_specific_kwargs = training_params.lr_scheduler_kwargs.copy()
    warmup_steps = training_params.warmup_steps
    warmup_ratio = training_params.warmup_ratio
    last_epoch = current_epoch - 1

    if warmup_steps is not None and warmup_ratio is not None:
        raise ValueError("Only one of warmup_steps and warmup_ratio should be provided")

    # Make sure num_training_steps is provided for schedulers that need it
    if (
        scheduler_type
        in (
            SchedulerType.COSINE,
            SchedulerType.COSINE_WITH_RESTARTS,
            SchedulerType.COSINE_WITH_MIN_LR,
            SchedulerType.LINEAR,
        )
        and num_training_steps is None
    ):
        raise ValueError(
            f"num_training_steps must be provided when using {scheduler_type}"
        )

    # Set warmup_steps based on warmup_ratio if provided
    if warmup_ratio is not None:
        if warmup_ratio < 0 or warmup_ratio > 1:
            raise ValueError("warmup_ratio must be in [0, 1]")

        if num_training_steps is None:
            raise ValueError(
                "num_training_steps must be provided when using warmup_ratio"
            )

        warmup_steps = int(warmup_ratio * num_training_steps)
        logger.info(
            f"Using warmup_steps={warmup_steps} based on "
            f"{warmup_ratio} warmup_ratio and {num_training_steps} max steps."
        )

    # Otherwise set warmup_steps to 0 if not provided
    if warmup_steps is None:
        warmup_steps = 0
        logger.info("No warmup steps provided. Setting warmup_steps=0.")

    if scheduler_type == SchedulerType.LINEAR:
        if scheduler_specific_kwargs:
            logger.warning(
                f"Unrecognized scheduler kwargs: {scheduler_specific_kwargs}. "
                "Ignoring them."
            )

        return get_linear_schedule_with_warmup(
            optimizer=optimizer,
            num_warmup_steps=warmup_steps,
            num_training_steps=num_training_steps,
            last_epoch=last_epoch,
        )
    elif scheduler_type == SchedulerType.COSINE:
        assert num_training_steps is not None, "should not be none by this point"

        num_cycles = scheduler_specific_kwargs.pop("num_cycles", 0.5)

        if scheduler_specific_kwargs:
            logger.warning(
                f"Unrecognized scheduler kwargs: {scheduler_specific_kwargs}. "
                "Ignoring them."
            )

        return get_cosine_schedule_with_warmup(
            optimizer=optimizer,
            num_warmup_steps=warmup_steps,
            num_training_steps=num_training_steps,
            last_epoch=last_epoch,
            num_cycles=num_cycles,
        )
    elif scheduler_type == SchedulerType.CONSTANT:
        if scheduler_specific_kwargs:
            logger.warning(
                f"Unrecognized scheduler kwargs: {scheduler_specific_kwargs}. "
                "Ignoring them."
            )

        return transformers.get_constant_schedule_with_warmup(
            optimizer=optimizer,
            num_warmup_steps=warmup_steps,
            last_epoch=last_epoch,
        )
    elif scheduler_type == SchedulerType.COSINE_WITH_RESTARTS:
        assert num_training_steps is not None, "should not be none by this point"

        num_cycles = scheduler_specific_kwargs.pop("num_cycles", 1)

        if scheduler_specific_kwargs:
            logger.warning(
                f"Unrecognized scheduler kwargs: {scheduler_specific_kwargs}. "
                "Ignoring them."
            )

        return get_cosine_with_hard_restarts_schedule_with_warmup(
            optimizer=optimizer,
            num_warmup_steps=warmup_steps,
            num_training_steps=num_training_steps,
            last_epoch=last_epoch,
            num_cycles=num_cycles,
        )
    elif scheduler_type == SchedulerType.COSINE_WITH_MIN_LR:
        assert num_training_steps is not None, "should not be none by this point"

        num_cycles = scheduler_specific_kwargs.pop("num_cycles", 0.5)
        min_lr_rate = scheduler_specific_kwargs.pop("min_lr_rate", 0.0)

        if scheduler_specific_kwargs:
            logger.warning(
                f"Unrecognized scheduler kwargs: {scheduler_specific_kwargs}. "
                "Ignoring them."
            )

        return transformers.optimization.get_cosine_with_min_lr_schedule_with_warmup(
            optimizer=optimizer,
            num_warmup_steps=warmup_steps,
            num_training_steps=num_training_steps,
            last_epoch=last_epoch,
            num_cycles=num_cycles,
            min_lr_rate=min_lr_rate,
        )
    else:
        raise ValueError(f"Unknown scheduler type: {scheduler_type}")



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oumi.inference.remote_inference_engine import RemoteInferenceEngine


class DeepSeekInferenceEngine(RemoteInferenceEngine):
    """Engine for running inference against the DeepSeek API.

    Documentation: https://api-docs.deepseek.com
    """

    base_url = "https://api.deepseek.com/v1/chat/completions"
    """The base URL for the DeepSeek API."""

    api_key_env_varname = "DEEPSEEK_API_KEY"
    """The environment variable name for the DeepSeek API key."""



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

from oumi.core.configs import EvaluationConfig
from oumi.core.evaluation.evaluation_result import EvaluationResult
from oumi.core.evaluation.evaluator import Evaluator


def evaluate(config: EvaluationConfig) -> list[dict[str, Any]]:
    """Evaluates a model using the provided configuration.

    Args:
        config: The desired configuration for evaluation.

    Returns:
        A list of evaluation results (one for each task). Each evaluation result is a
        dictionary of metric names and their corresponding values.
    """
    evaluator = Evaluator()
    results: list[EvaluationResult] = evaluator.evaluate(config)
    return [result.task_result for result in results]



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import math
import time
from collections import deque

from oumi.core.configs.params.remote_params import AdaptiveConcurrencyParams
from oumi.inference.adaptive_semaphore import PoliteAdaptiveSemaphore
from oumi.utils.logging import logger

_MAX_OUTCOMES_WINDOW_SIZE = 1000


class AdaptiveConcurrencyController:
    """Controls concurrency based on error rates.

    The controller functions as a semaphore, but can dynamically adjust the number of
    concurrent requests based on error rates by recording the outcomes of requests.

    The controller will not adjust the concurrency unless the number of recent requests
    is greater than or equal to the min_window_size, and the time since the last
    adjustment is greater than or equal to the min_update_time.

    To avoid making adjustments on stale data, the controller clears the outcomes
    queue whenever the concurrency is adjusted. In addition, the controller waits for
    consecutive good windows before recovering from backoff, and consecutive error
    windows before triggering additional backoff.

    To use the controller, use it as a semaphore while recording request outcomes.

    Example:
    ```python
    controller = AdaptiveConcurrencyController(config)
    async with controller.acquire():
        try:
            # Send async request
            response = await send_async_request()
            controller.record_success()
        except Exception as e:
            controller.record_error()
    ```
    """

    def __init__(self, config: AdaptiveConcurrencyParams, politeness_policy: float):
        """Initialize the adaptive concurrency controller.

        Args:
            config: Configuration for adaptive concurrency control.
            politeness_policy: The amount of time workers will wait until sending their
                next request, in seconds.
        """
        self._config = config
        self._current_concurrency = self._get_initial_concurrency()
        self._semaphore = PoliteAdaptiveSemaphore(
            self._current_concurrency, politeness_policy
        )

        # Track request outcomes in a sliding window
        self._outcomes: deque[bool] = deque(maxlen=_MAX_OUTCOMES_WINDOW_SIZE)
        self._outcome_lock = asyncio.Lock()

        # State tracking
        self._last_adjustment_time = 0
        self._last_warmup_time = 0
        self._in_backoff = False
        self._consecutive_good_windows_since_last_update = 0
        self._consecutive_error_windows_since_last_update = 0
        self._consecutive_good_windows_required_for_recovery = 2
        self._consecutive_error_windows_for_additional_backoff = 2

        self._logged_backoff_warning = False
        self._logged_warmup_warning = False

    async def record_success(self):
        """Record a successful outcome."""
        async with self._outcome_lock:
            self._outcomes.append(True)

    async def record_error(self):
        """Record a failed outcome."""
        async with self._outcome_lock:
            self._outcomes.append(False)

    async def acquire(self):
        """Acquire a permit, then try to adjust concurrency before returning."""
        await self._semaphore.acquire()
        await self._try_adjust_concurrency()

    def release(self):
        """Release a permit."""
        self._semaphore.release()

    async def __aenter__(self):
        """Enter the context manager."""
        await self.acquire()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        """Exit the context manager."""
        self.release()

    def _get_initial_concurrency(self) -> int:
        """Get the initial concurrency."""
        potential_concurrency = math.floor(
            (self._config.max_concurrency - self._config.min_concurrency)
            * self._config.initial_concurrency_factor
            + self._config.min_concurrency
        )
        # Ensure the initial concurrency is at least the min concurrency and no greater
        # than the max concurrency.
        return max(
            self._config.min_concurrency,
            min(potential_concurrency, self._config.max_concurrency),
        )

    async def _get_error_rate(self) -> float:
        """Calculate current error rate based on recent outcomes."""
        async with self._outcome_lock:
            if not self._outcomes:
                return 0.0  # No data yet, so no error rate.
            errors = sum(1 for outcome in self._outcomes if not outcome)
            return errors / len(self._outcomes)

    async def _try_adjust_concurrency(self) -> None:
        """Try to adjust concurrency based on current error rate and state."""
        # No data yet, so no information to adjust concurrency.
        if not self._outcomes or len(self._outcomes) < self._config.min_window_size:
            return

        # If we haven't reached the update time, do nothing
        time_since_last_adjustment = time.time() - self._last_adjustment_time
        if time_since_last_adjustment < self._config.min_update_time:
            return

        # Check if we need to backoff due to high error rate
        error_rate = await self._get_error_rate()
        if error_rate >= self._config.error_threshold and not self._in_backoff:
            await self._try_backoff()
            return

        # Check if we can recover from backoff
        if self._in_backoff:
            if error_rate <= self._config.recovery_threshold:
                self._consecutive_good_windows_since_last_update += 1
                self._consecutive_error_windows_since_last_update = 0
                await self._reset_outcomes()
                # Require multiple good windows before recovering
                if (
                    self._consecutive_good_windows_since_last_update
                    >= self._consecutive_good_windows_required_for_recovery
                ):
                    self._end_backoff()
                    # Don't return because we want to try to warmup
                else:
                    return
            else:
                self._consecutive_error_windows_since_last_update += 1
                self._consecutive_good_windows_since_last_update = 0
                await self._reset_outcomes()
                if (
                    self._consecutive_error_windows_since_last_update
                    >= self._consecutive_error_windows_for_additional_backoff
                ):
                    await self._try_backoff()
                return

        await self._try_warmup()

    def _end_backoff(self):
        """End backoff state."""
        self._in_backoff = False
        self._consecutive_good_windows_since_last_update = 0
        self._consecutive_error_windows_since_last_update = 0

    async def _try_backoff(self):
        """Try to reduce concurrency due to high error rate."""
        new_concurrency = max(
            self._config.min_concurrency,
            # Round down to nearest integer
            math.floor(self._current_concurrency * self._config.backoff_factor),
        )

        if new_concurrency != self._current_concurrency:
            logger.info(
                "Lowering concurrency from %d to %d due to high error rate.",
                self._current_concurrency,
                new_concurrency,
            )
            await self._update_concurrency(new_concurrency)
        else:
            if not self._logged_backoff_warning:
                logger.warning(
                    "Entering backoff state, but concurrency is already at minimum "
                    "value. Consider lowering the min concurrency or increasing the "
                    "time between requests (politeness_policy)."
                )
                self._logged_backoff_warning = True

        # Set backoff state
        self._in_backoff = True

    async def _reset_outcomes(self):
        """Reset the outcomes queue."""
        async with self._outcome_lock:
            self._outcomes.clear()

    async def _clear_adjustment_state(self):
        """Reset adjustment state variables."""
        self._consecutive_good_windows_since_last_update = 0
        self._consecutive_error_windows_since_last_update = 0
        self._last_adjustment_time = time.time()

    async def _try_warmup(self):
        """Try to increase concurrency during warmup."""
        new_concurrency = min(
            self._config.max_concurrency,
            self._current_concurrency + self._config.concurrency_step,
        )

        if new_concurrency != self._current_concurrency:
            await self._update_concurrency(new_concurrency)
        else:
            if not self._logged_warmup_warning:
                logger.warning(
                    "Entering warmup state, but concurrency is already at maximum "
                    "value. Consider raising the max concurrency."
                )
                self._logged_warmup_warning = True

    async def _update_concurrency(self, new_concurrency: int):
        """Update the concurrency limit, preserving existing waiters."""
        self._current_concurrency = new_concurrency
        await self._semaphore.adjust_capacity(new_concurrency)
        await self._reset_outcomes()
        await self._clear_adjustment_state()



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import copy
import math
import os
import time
from contextlib import contextmanager
from pathlib import Path
from pprint import pformat
from typing import Any, Callable, Optional, cast

import mlflow
import pydantic
import safetensors.torch
import torch
import torch.amp
import torch.distributed.checkpoint as dcp
import torch.utils.tensorboard as tensorboard
import wandb
from torch.distributed.checkpoint.state_dict import (
    StateDictOptions,
    get_state_dict,
)
from torch.utils.data import DataLoader, Dataset, DistributedSampler, IterableDataset
from torchdata.stateful_dataloader import StatefulDataLoader
from tqdm.auto import tqdm
from transformers import TrainerCallback

from oumi.core.configs import MixedPrecisionDtype, TrainingConfig, TrainingParams
from oumi.core.configs.params.fsdp_params import FSDPParams, StateDictType
from oumi.core.distributed import (
    barrier,
    get_device_rank_info,
    is_distributed,
    is_local_process_zero,
    is_world_process_zero,
    prepare_model_for_distributed,
)
from oumi.core.processors.base_processor import BaseProcessor
from oumi.core.tokenizers import BaseTokenizer
from oumi.core.trainers.base_trainer import BaseTrainer
from oumi.performance.telemetry import TelemetryTracker
from oumi.utils.io_utils import load_json, save_json
from oumi.utils.logging import logger
from oumi.utils.serialization_utils import flatten_config


class TrainingState(pydantic.BaseModel):
    epoch: int = 0
    global_step: int = 0
    total_tokens_seen: int = 0


class Trainer(BaseTrainer):
    def __init__(
        self,
        model: torch.nn.Module,
        processing_class: Optional[BaseTokenizer],
        args: TrainingParams,
        train_dataset: Dataset,
        processor: Optional[BaseProcessor] = None,
        eval_dataset: Optional[Dataset] = None,
        callbacks: Optional[list[TrainerCallback]] = None,
        data_collator: Optional[Callable] = None,
        config: Optional[TrainingConfig] = None,
        **kwargs,
    ):
        """Initializes the Oumi trainer."""
        # Importing these here to avoid circular dependencies
        from oumi.builders.lr_schedules import build_lr_scheduler
        from oumi.builders.optimizers import build_optimizer

        torch.backends.cuda.matmul.allow_tf32 = True  # allow tf32 on matmul
        torch.backends.cudnn.allow_tf32 = True  # allow tf32 on cudnn

        self.telemetry = TelemetryTracker()
        self.start_time = time.perf_counter()
        self.collator_fn = data_collator

        self.processing_class = processing_class
        self._processor = processor
        self.params = copy.deepcopy(args)
        self.train_dataset = train_dataset
        self.eval_dataset = eval_dataset
        self.max_norm = (
            float(args.max_grad_norm) if args.max_grad_norm is not None else None
        )

        self.config = config or TrainingConfig()
        self.fsdp_params = self.config.fsdp or FSDPParams()
        self.is_using_fsdp = self.fsdp_params.enable_fsdp
        # TODO OPE-333 Define a param to enable ring attention + check pre-conditions:
        # 1. Flash Attention (`is_ring_attention_available()`),
        # 2. CUDA and distributed multi-GPU training (otherwise, pointless).
        # 3. Supported model type.
        self.is_using_ring_attention = False
        self._mlflow_oumi_managed_run = (
            # If the user has manually started an mlflow run, we don't need to start
            # a new one and can let the user manage it themselves.
            # Otherwise, we start a new run and end it when training is done.
            self.params.enable_mlflow and not mlflow.active_run()
        )

        self.params.finalize_and_validate()

        self.state = TrainingState()
        self.device_type = "cuda" if torch.cuda.is_available() else "cpu"

        # Enable mixed precision bf16/fp16 training if requested.
        # Model dtype has been verified to be fp32 if this is the case.
        self.mixed_precision_ctx = contextlib.nullcontext()
        mixed_precision_dtype = None
        if self.params.mixed_precision_dtype == MixedPrecisionDtype.BF16:
            mixed_precision_dtype = torch.bfloat16
        elif self.params.mixed_precision_dtype == MixedPrecisionDtype.FP16:
            mixed_precision_dtype = torch.float16
        if mixed_precision_dtype:
            self.mixed_precision_ctx = torch.amp.autocast(
                device_type=self.device_type,
                enabled=True,
                dtype=mixed_precision_dtype,
            )

        # We want to enable gradient scaling for fp16 mixed precision training
        # to prevent gradient underflows. This is not needed for bf16 since it has the
        # same dynamic range as fp32. See here for details:
        # https://pytorch.org/docs/stable/amp.html#gradient-scaling
        self.scaler = torch.amp.GradScaler(
            device=self.device_type,
            enabled=self.params.mixed_precision_dtype == MixedPrecisionDtype.FP16,
        )

        device_info = get_device_rank_info()

        # TODO: OPE-218 - give users fine-grained control on device placement
        # TODO: OPE-217 - non-leader models should be on meta
        if torch.cuda.is_available():
            self.device = f"cuda:{device_info.local_rank}"
            torch.cuda.set_device(self.device)
        elif torch.backends.mps.is_available():
            self.device = "mps"
        else:
            self.device = "cpu"

        # ----------------------------------
        # Prepare model for training
        # ----------------------------------
        if args.enable_gradient_checkpointing:
            if not hasattr(model, "gradient_checkpointing_enable"):
                raise ValueError(
                    "Gradient checkpointing is only supported for Hugging Face models."
                )
            model.gradient_checkpointing_enable(args.gradient_checkpointing_kwargs)  # type: ignore
        model = cast(torch.nn.Module, model)
        model.to(self.device)
        if is_distributed():
            # Wrap model for distributed training
            with self._telemetry_block("wrap model for distributed"):
                model = prepare_model_for_distributed(
                    model,
                    self.config,
                    ddp_find_unused_parameters=self.params.ddp_find_unused_parameters,
                )

        if self.params.compile:
            self.log("Compiling model...")
            with self._telemetry_block("compile model"):
                model = cast(torch.nn.Module, torch.compile(model))
        self.model = model

        self.callbacks = callbacks if callbacks is not None else []

        self.optimizer = build_optimizer(self.model, self.params)
        self.lr_scheduler = build_lr_scheduler(
            optimizer=self.optimizer,
            training_params=self.params,
            current_epoch=self.state.epoch,
            num_training_steps=self._estimate_total_training_steps(),
        )

        self.train_dataloader = self._get_train_dataloader()
        self.eval_dataloader = self._get_eval_dataloader() if eval_dataset else None

        self._init_logging()

    #
    # Training
    #
    def train(self, resume_from_checkpoint: Optional[str] = None):
        """Trains the model."""
        if resume_from_checkpoint:
            with torch.profiler.record_function("load_from_checkpoint"):
                self._load_from_checkpoint(resume_from_checkpoint)
        else:
            # Log training config and parameters to all enabled platforms
            self._log_training_config()

        total_steps = self._estimate_total_training_steps()

        self.start_time = time.perf_counter()

        # Make sure all workers start at the same time.
        barrier()

        with tqdm(
            total=total_steps,
            desc="Training",
            disable=not is_world_process_zero(),
        ) as progress_bar:
            while True:
                epoch = self.state.epoch
                if self.params.max_steps > 0:
                    if self.state.global_step >= self.params.max_steps:
                        self.log(
                            f"Reached {self.state.global_step} global steps. "
                            "Training completed."
                        )
                        break
                elif (
                    self.params.num_train_epochs > 0
                    and epoch >= self.params.num_train_epochs
                ):
                    self.log(f"Reached {epoch} epochs. Training completed.")
                    break

                with torch.profiler.record_function(f"epoch_{epoch}"):
                    self._set_sampler_epoch(epoch)
                    self._train_epoch(progress_bar)

                    if self.params.save_epoch:
                        self.save_state()

                    if (
                        self.eval_dataloader
                        and self.params.eval_strategy == "epoch"
                        and is_world_process_zero()
                    ):
                        # TODO: OPE-223 - only the global leader is used for evaluation
                        # To enable distributed evaluation, the eval function needs
                        # to be updated to aggregate metrics accross all workers.
                        self.evaluate()

                    self.state.epoch += 1

                    barrier()

            self._process_callbacks("on_train_end")

        self.log(
            f"Training finished! Global step: {self.state.global_step} "
            f"Training runtime: {time.perf_counter() - self.start_time}s"
        )

        if self.params.enable_mlflow and not self._mlflow_oumi_managed_run:
            mlflow.end_run()

    @contextmanager
    def _telemetry_block(self, name: str):
        with (
            torch.profiler.record_function(name) as record_function_context,
            self.telemetry.timer(name) as timer_context,
        ):
            yield (record_function_context, timer_context)

    @staticmethod
    def _cuda_sync_and_empty_cache() -> None:
        if torch.cuda.is_available() and torch.cuda.is_initialized():
            torch.cuda.synchronize()
            torch.cuda.empty_cache()

    def _train_epoch(self, progress_bar: tqdm) -> None:
        """Trains the model for one epoch."""
        epoch_start_time = time.perf_counter()

        self.model.train()
        self._cuda_sync_and_empty_cache()
        self.optimizer.zero_grad(set_to_none=True)
        micro_step = 0

        data_iter = iter(self.train_dataloader)

        gradient_accumulation_steps = max(1, self.params.gradient_accumulation_steps)

        while True:
            with torch.profiler.record_function(
                "microstep" if gradient_accumulation_steps > 1 else "step"
            ):
                if micro_step % gradient_accumulation_steps == 0:
                    self._process_callbacks("on_step_begin")

                # True if `max_steps` is configured and we reached the limit.
                stop_on_max_steps_limit = (
                    self.params.max_steps > 0
                    and (self.state.global_step + 1) >= self.params.max_steps
                )
                # End of global step. May include multiple micro steps
                # if gradient_accumulation_steps > 1.
                end_of_global_step = (
                    (micro_step + 1) % gradient_accumulation_steps
                ) == 0

                with self._telemetry_block("fetching batch"):
                    try:
                        batch = next(data_iter)
                    except StopIteration:
                        # FIXME Update metrics and log
                        self.log("End of epoch")
                        break

                # Count tokens on CPU.
                with self._telemetry_block("computing tokens"):
                    if self.processing_class is not None and "input_ids" in batch:
                        num_tokens = (
                            batch["input_ids"]
                            .ne(self.processing_class.pad_token_id)
                            .sum()
                            .item()
                        )
                        self.state.total_tokens_seen += num_tokens

                with self._telemetry_block("moving batch to device"):
                    if not self.is_using_fsdp:
                        batch = {
                            k: v.to(self.device, non_blocking=True)
                            for k, v in batch.items()
                        }

                with self.mixed_precision_ctx, self._telemetry_block("model forward"):
                    self.model.require_backward_grad_sync = (  # type: ignore
                        end_of_global_step or stop_on_max_steps_limit
                    )

                    outputs = self.model(**batch)

                    loss = outputs["loss"] / gradient_accumulation_steps

                with self._telemetry_block("loss backward"):
                    self.scaler.scale(loss).backward()

                if end_of_global_step or stop_on_max_steps_limit:
                    with self._telemetry_block("optimizer step"):
                        self.scaler.unscale_(self.optimizer)
                        if self.max_norm is not None and self.max_norm > 0:
                            torch.nn.utils.clip_grad_norm_(
                                self.model.parameters(), max_norm=self.max_norm
                            )

                        # save lr for logging
                        last_lr = self.lr_scheduler.get_last_lr()[0]

                        # step optimizer, scaler, and lr schedule
                        self.scaler.step(self.optimizer)
                        self.scaler.update()
                        self.lr_scheduler.step()

                        self.optimizer.zero_grad(set_to_none=True)

                    self.state.global_step += 1
                    if self.params.telemetry.track_gpu_temperature:
                        self.telemetry.record_gpu_temperature()
                    progress_bar.update(1)

                    self._process_callbacks("on_step_end")

                    if (
                        self.params.logging_steps > 0
                        and not (
                            self.state.global_step == 1
                            and self.params.logging_first_step
                        )
                        and (
                            stop_on_max_steps_limit
                            or (self.state.global_step % self.params.logging_steps == 0)
                        )
                    ):
                        # Log metrics
                        elapsed = time.perf_counter() - self.start_time
                        loss_value = loss.item() * gradient_accumulation_steps
                        metrics = {
                            "train/loss": loss_value,
                            "learning_rate": last_lr,
                            "epoch": self.state.epoch,
                            "global_step": self.state.global_step,
                            "total_tokens_seen": self.state.total_tokens_seen,
                            "global_steps_per_second": self.state.global_step / elapsed,
                            "tokens_per_second": self.state.total_tokens_seen / elapsed,
                            "tokens_per_step_per_gpu": self.state.total_tokens_seen
                            / self.state.global_step,
                        }
                        callback_metrics = self._process_callbacks("on_log", metrics)
                        metrics.update(callback_metrics)

                        self.log_metrics(metrics, self.state.global_step)

                        if is_local_process_zero():
                            self.telemetry.print_summary()

                    if (
                        self.params.save_steps > 0
                        and self.state.global_step % self.params.save_steps == 0
                    ):
                        self.save_state()

                    if (
                        self.eval_dataloader
                        and self.params.eval_steps > 0
                        and self.state.global_step % self.params.eval_steps == 0
                        and is_world_process_zero()
                    ):
                        # TODO: OPE-223 - only the global leader is used for evaluation
                        # To enable distributed evaluation, th eval function needs
                        # to be updated to aggregate metrics accross all workers.
                        self.evaluate()

                if stop_on_max_steps_limit:
                    self.log(f"Reached {self.params.max_steps} max steps condition.")
                    break

                micro_step += 1

        self.log(
            f"End of epoch. "
            f"Global step: {self.state.global_step}. "
            f"Epoch runtime: {time.perf_counter() - epoch_start_time}s"
        )

    #
    # Evaluation
    #
    @torch.no_grad()
    def evaluate(self) -> dict[str, float]:
        """Evaluates the model on the evaluation dataset."""
        if self.eval_dataloader is None:
            raise ValueError("No evaluation dataloader provided.")

        self.model.eval()
        eval_losses = []

        for batch in tqdm(
            self.eval_dataloader,
            desc="Evaluating",
            disable=not is_local_process_zero(),
        ):
            batch = {k: v.to(self.device) for k, v in batch.items()}
            outputs = self.model(**batch)
            eval_losses.append(outputs.loss.item())

        eval_loss = sum(eval_losses) / len(eval_losses)
        perplexity = torch.exp(torch.tensor(eval_loss))

        results = {"val/loss": eval_loss, "val/perplexity": perplexity.item()}

        self.log("Finished evaluation.")
        self.log_metrics(results, self.state.global_step)

        self.model.train()
        return results

    #
    # Checkpointing
    #
    def save_model(self, config: TrainingConfig, final: bool = True) -> None:
        """Saves the model."""
        self._cuda_sync_and_empty_cache()
        if is_world_process_zero():
            output_dir = Path(config.training.output_dir)
            output_dir.mkdir(exist_ok=True)
            model_path = output_dir / "model.safetensors"
            safetensors.torch.save_model(model=self.model, filename=str(model_path))
            self.log(f"Model saved to {model_path}.")

            if self._processor is not None:
                self._processor.save_config(output_dir)
                logger.info(f"Processor config has been saved at {output_dir}.")
        self._cuda_sync_and_empty_cache()

    def save_state(self):
        """Saves the training state."""
        self._cuda_sync_and_empty_cache()
        checkpoint_dir = Path(self.params.output_dir)

        if is_local_process_zero():
            checkpoint_dir.mkdir(exist_ok=True)

        if (
            self.params.telemetry.collect_telemetry_for_all_ranks
            or is_world_process_zero()
        ):
            telemetry_dir = self.params.telemetry_dir
            if telemetry_dir:
                device_rank_info = get_device_rank_info()
                telemetry_state_path = (
                    telemetry_dir / f"telemetry_rank{device_rank_info.rank:04}.json"
                )
                save_json(
                    data=self.telemetry.state_dict(),
                    filename=telemetry_state_path,
                )

        if self.is_using_fsdp:
            storage_options = StateDictOptions(
                full_state_dict=self.fsdp_params.state_dict_type
                == StateDictType.FULL_STATE_DICT,
                cpu_offload=self.fsdp_params.cpu_offload,
                ignore_frozen_params=False,
                strict=True,
                broadcast_from_rank0=False,  # TODO: make this configurable
            )
        else:
            storage_options = None

        model_state_dict, optimizer_state_dict = get_state_dict(
            model=self.model,
            optimizers=self.optimizer,
            options=storage_options,
        )

        model_path = checkpoint_dir / "model"
        optimizer_path = checkpoint_dir / "optimizer"
        dataloader_state_path = checkpoint_dir / "dataloader.pt"
        trainer_state_path = checkpoint_dir / "trainer_state.json"

        dcp.save(model_state_dict, checkpoint_id=model_path)
        dcp.save(optimizer_state_dict, checkpoint_id=optimizer_path)

        if is_world_process_zero():
            torch.save(self.train_dataloader.state_dict(), dataloader_state_path)
            save_json(data=self.state.model_dump(), filename=trainer_state_path)
            logger.info(f"Training state saved to {checkpoint_dir}")

        self._cuda_sync_and_empty_cache()

    def _load_from_checkpoint(self, checkpoint_dirname: str):
        """Loads the training state from a checkpoint."""
        checkpoint_dir = Path(checkpoint_dirname)

        device_rank_info = get_device_rank_info()

        model_path = checkpoint_dir / "model"
        optimizer_path = checkpoint_dir / "optimizer"
        dataloader_state_path = checkpoint_dir / "dataloader.pt"
        trainer_state_path = checkpoint_dir / "trainer_state.json"
        telemetry_state_path = (
            checkpoint_dir / f"telemetry_rank{device_rank_info.rank:04}.json"
        )

        if not checkpoint_dir.exists():
            raise ValueError(f"Checkpoint directory does not exist: {checkpoint_dir}")
        if not model_path.exists():
            raise ValueError(
                f"Invalid checkpoint, model state folder does not exist: {model_path}"
            )
        if not optimizer_path.exists():
            raise ValueError(
                "Invalid checkpoint, optimizer state folder does not exist: "
                f"{optimizer_path}"
            )

        if self.is_using_fsdp:
            storage_options = StateDictOptions(
                full_state_dict=self.fsdp_params.state_dict_type
                == StateDictType.FULL_STATE_DICT,
                cpu_offload=self.fsdp_params.cpu_offload,
                ignore_frozen_params=False,
                strict=True,
                broadcast_from_rank0=False,
            )
        else:
            storage_options = None

        model_state_dict, optimizer_state_dict = get_state_dict(
            model=self.model,
            optimizers=self.optimizer,
            options=storage_options,
        )

        dcp.load(model_state_dict, checkpoint_id=model_path)
        dcp.load(optimizer_state_dict, checkpoint_id=optimizer_path)

        if dataloader_state_path.exists():
            self.train_dataloader.load_state_dict(torch.load(dataloader_state_path))
        if trainer_state_path.exists():
            self.state = TrainingState.model_validate(
                load_json(trainer_state_path), strict=True
            )
        if telemetry_state_path.exists():
            self.telemetry.load_state_dict(load_json(telemetry_state_path))

        self.log(f"Resumed training from checkpoint: {checkpoint_dirname}")

    #
    # Logging
    #
    def log(self, message: str):
        """Logs a message if the process is the local process zero."""
        if not is_local_process_zero():
            return
        logger.info(message)

    def log_metrics(self, metrics: dict[str, Any], step: int) -> None:
        """Logs metrics to wandb and tensorboard."""
        # Log to console and log file
        if not is_world_process_zero():
            return
        self.log(pformat(metrics))

        # Log to Weights and Biases
        if self.params.enable_wandb:
            wandb.log(metrics, step=self.state.global_step)

        # Log to tensorboard
        if self.params.enable_tensorboard and self.tensorboard_writer:
            for key, value in metrics.items():
                self.tensorboard_writer.add_scalar(key, value, self.state.global_step)

        # Log to mlflow
        if self.params.enable_mlflow:
            mlflow.log_metrics(metrics, step=self.state.global_step)

    def _init_logging(
        self,
    ) -> None:
        """Initializes logging."""
        if not is_world_process_zero():
            return

        self.log(f"Logging to {self.params.output_dir}")

        if self.params.enable_wandb:
            project_name = os.environ.get("WANDB_PROJECT", "oumi")
            self.log(f"Logging to Weights and Biases project: '{project_name}'")
            run = wandb.init(
                project=project_name, name=self.params.run_name, job_type="train"
            )
            self.log(f"View wandb run {run.id} at: {run.get_url()}")
            wandb.watch(self.model)

        if self.params.enable_tensorboard:
            tensorboard_folder = Path(self.params.output_dir) / "tensorboard"
            self.log(f"Logging to tensorboard folder: '{tensorboard_folder}'")
            self.tensorboard_writer = tensorboard.SummaryWriter(
                log_dir=tensorboard_folder
            )
        else:
            self.tensorboard_writer = None

        if self.params.enable_mlflow and self._mlflow_oumi_managed_run:
            self.mlflow_run = mlflow.start_run(run_name=self.params.run_name)

    def _log_training_config(self) -> None:
        """Logs training configuration and parameters to all enabled platforms."""
        if not is_world_process_zero() or not self.config:
            return

        # Get flattened config from both training config and parameters
        config_dict = flatten_config(self.config)

        # Log to MLflow
        if self.params.enable_mlflow:
            try:
                mlflow.log_params(config_dict)
            except Exception as e:
                self.log(f"Failed to log config to MLflow: {e}")

        # Log to Weights & Biases
        if self.params.enable_wandb:
            try:
                # wandb.config can handle nested dictionaries, but we'll use
                # flattened for consistency
                wandb.config.update(config_dict)
            except Exception as e:
                self.log(f"Failed to log config to wandb: {e}")

        # Log to TensorBoard
        if self.params.enable_tensorboard and self.tensorboard_writer:
            try:
                # Log config as a formatted text to tensorboard
                config_text = "\n".join([f"{k}: {v}" for k, v in config_dict.items()])
                self.tensorboard_writer.add_text(
                    tag="config/training_config",
                    text_string=config_text,
                    global_step=self.state.global_step,
                )
            except Exception as e:
                self.log(f"Failed to log config to TensorBoard: {e}")

    #
    # Data loading
    #
    def _get_train_dataloader(self) -> StatefulDataLoader:
        """Returns the training dataloader."""
        # At this point, "auto" must be pre-resolved to `int`.
        assert isinstance(self.params.dataloader_num_workers, int)
        prefetch_factor = (
            self.params.dataloader_prefetch_factor
            if self.params.dataloader_num_workers > 0
            else None
        )

        # IterDataPipe is a subclass of IterableDataset.
        if isinstance(self.train_dataset, IterableDataset):
            # TODO: configure sharding for iterable datasets
            sampler = None
            shuffle = None
        else:
            # Configure sampler for map datasets. If using multiple GPUs,
            # we use a DistributedSampler to make sure each worker gets a
            # different subset of the dataset.
            # In non-distributed mode, we iterate over the full dataset.
            if is_distributed():
                # TODO: OPE-219 this strategy should only be enabled for DDP
                # and FSDP with NO_SHARDING
                device_info = get_device_rank_info()

                # Distribute the dataset across all GPU workers
                # Each rank will get a subset of the dataset
                sampler = DistributedSampler(
                    self.train_dataset,
                    num_replicas=device_info.world_size,
                    rank=device_info.rank,
                    seed=self.params.seed,
                    shuffle=True,
                )
                shuffle = False
            else:
                # If not distributed, let the dataloader handle shuffling
                sampler = None
                shuffle = True

        # Keeping track of the sampler so we can update after each epoch
        self._sampler = sampler

        return StatefulDataLoader(
            self.train_dataset,
            batch_size=self.params.per_device_train_batch_size,
            shuffle=shuffle,
            sampler=self._sampler,
            num_workers=self.params.dataloader_num_workers,
            pin_memory=self.device_type == "cuda",
            prefetch_factor=prefetch_factor,
            pin_memory_device=self.device,
            snapshot_every_n_steps=self.params.save_steps,
            collate_fn=self.collator_fn,
        )

    def _get_eval_dataloader(self) -> DataLoader:
        """Returns the evaluation dataloader."""
        if not self.eval_dataset:
            raise ValueError("No evaluation dataset provided.")

        # At this point, "auto" must be pre-resolved to `int`.
        assert isinstance(self.params.dataloader_num_workers, int)
        return DataLoader(
            self.eval_dataset,
            batch_size=self.params.per_device_eval_batch_size,
            shuffle=False,
            num_workers=self.params.dataloader_num_workers,
            collate_fn=self.collator_fn,
        )

    def _estimate_total_training_steps(self) -> int:
        # If max_steps is set, use it.
        if self.params.max_steps > 0:
            return self.params.max_steps

        num_epochs = self.params.num_train_epochs
        if num_epochs > 0:
            num_dataset_examples = 0
            try:
                if not isinstance(self.train_dataset, IterableDataset):
                    num_dataset_examples = len(self.train_dataset)  # type: ignore
                elif hasattr(self.train_dataset, "datapipe"):
                    # Hacky way to get examples count from
                    # MapToIterConverterIterDataPipe.
                    # FIXME Remove DataPipes OPE-811
                    num_dataset_examples = len(self.train_dataset.datapipe)  # type: ignore
            except Exception:
                num_dataset_examples = 0

            if num_dataset_examples > 0:
                world_size = get_device_rank_info().world_size
                batch_size = self.params.per_device_train_batch_size
                steps_per_epoch_per_device = math.ceil(
                    float(num_dataset_examples) / (batch_size * world_size)
                )
                return int(num_epochs * max(steps_per_epoch_per_device, 1))

        raise ValueError(
            "Unable to estimate `total_training_steps` "
            + (f"in {num_epochs} epochs" if num_epochs > 0 else "")
            + ". Please define `max_steps` training parameter!"
        )

    def _set_sampler_epoch(self, epoch: int) -> None:
        """Sets the current epoch on sampler, if it exists and supports it."""
        if self._sampler and hasattr(self._sampler, "set_epoch"):
            self.log(f"Setting sampler epoch to {epoch}.")
            self._sampler.set_epoch(epoch)

    #
    # Handle callbacks
    #
    def _process_callbacks(
        self, event: str, logs: Optional[dict[str, Any]] = None
    ) -> dict[str, Any]:
        """Process callbacks.

        Extremely hacky way to handle HF callbacks.
        Just here to unblock debugging with our MfuCallback
        """
        logs = logs or {}

        for callback in self.callbacks:
            if hasattr(callback, event):
                action = getattr(callback, event)
                action(args=self.params, state=None, control=None, logs=logs)

        return logs

    def get_last_eval_metrics(self) -> dict[str, Any]:
        """TODO: This should be implemented in the future to enavle Oumi Trainer in the hyper parameters tuning."""  # noqa: E501
        raise NotImplementedError



from unittest.mock import Mock, patch

import pytest
import sky

from oumi.core.configs import JobConfig, JobResources, StorageMount
from oumi.core.launcher import JobState, JobStatus
from oumi.core.registry import REGISTRY, RegistryType
from oumi.launcher.clients.sky_client import SkyClient
from oumi.launcher.clouds.sky_cloud import SkyCloud
from oumi.launcher.clusters.sky_cluster import SkyCluster


#
# Fixtures
#
@pytest.fixture
def mock_sky_client():
    with patch("oumi.launcher.clouds.sky_cloud.SkyClient") as client:
        client.SupportedClouds = SkyClient.SupportedClouds
        client_instance = Mock(spec=SkyClient)
        client.return_value = client_instance
        yield client_instance


@pytest.fixture
def mock_sky_cluster():
    with patch("oumi.launcher.clouds.sky_cloud.SkyCluster") as cluster:
        yield cluster


def _get_default_job(cloud: str) -> JobConfig:
    resources = JobResources(
        cloud=cloud,
        region="us-central1",
        zone=None,
        accelerators="A100-80",
        cpus="4",
        memory="64",
        instance_type=None,
        use_spot=True,
        disk_size=512,
        disk_tier="low",
    )
    return JobConfig(
        name="myjob",
        user="user",
        working_dir="./",
        num_nodes=2,
        resources=resources,
        envs={"var1": "val1"},
        file_mounts={},
        storage_mounts={
            "~/home/remote/path/gcs/": StorageMount(
                source="gs://mybucket/", store="gcs"
            )
        },
        setup="pip install -r requirements.txt",
        run="./hello_world.sh",
    )


#
# Tests
#
def test_sky_cloud_up_cluster(mock_sky_client, mock_sky_cluster):
    expected_job_status = JobStatus(
        name="",
        id="1",
        cluster="new_cluster_name",
        status="",
        metadata="",
        done=False,
        state=JobState.PENDING,
    )
    mock_gcp_cluster = Mock(spec=sky.clouds.GCP)
    mock_gcp_handler = Mock()
    mock_gcp_handler.launched_resources = Mock()
    mock_gcp_handler.launched_resources.cloud = mock_gcp_cluster
    mock_sky_gcp = Mock(spec=SkyCluster)
    mock_sky_gcp.name.return_value = "new_cluster_name"
    mock_sky_gcp.get_job.return_value = expected_job_status

    mock_runpod_cluster = Mock(spec=sky.clouds.RunPod)
    mock_runpod_handler = Mock()
    mock_runpod_handler.launched_resources = Mock()
    mock_runpod_handler.launched_resources.cloud = mock_runpod_cluster
    mock_sky_runpod = Mock(spec=SkyCluster)
    mock_sky_runpod.name.return_value = "runpod_cluster"

    mock_lambda_cluster = Mock(spec=sky.clouds.Lambda)
    mock_lambda_handler = Mock()
    mock_lambda_handler.launched_resources = Mock()
    mock_lambda_handler.launched_resources.cloud = mock_lambda_cluster
    mock_sky_lambda = Mock(spec=SkyCluster)
    mock_sky_lambda.name.return_value = "lambda_cluster"
    mock_sky_cluster.side_effect = [
        mock_sky_gcp,
        mock_sky_runpod,
        mock_sky_lambda,
    ]
    mock_sky_client.status.return_value = [
        {
            "name": "new_cluster_name",
            "status": sky.ClusterStatus.UP,
            "handle": mock_gcp_handler,
        },
        {
            "name": "stop_cluster_name",
            "status": sky.ClusterStatus.STOPPED,
            "handle": mock_gcp_handler,
        },
        {
            "name": "init_cluster_name",
            "status": sky.ClusterStatus.INIT,
            "handle": mock_gcp_handler,
        },
        {
            "name": "runpod_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_runpod_handler,
        },
        {
            "name": "lambda_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_lambda_handler,
        },
    ]
    mock_sky_client.launch.return_value = expected_job_status
    cloud = SkyCloud("gcp")
    job_status = cloud.up_cluster(_get_default_job("gcp"), "new_cluster_name")
    mock_sky_client.launch.assert_called_once_with(
        _get_default_job("gcp"), "new_cluster_name"
    )
    assert job_status == expected_job_status


def test_sky_cloud_up_cluster_kwargs(mock_sky_client, mock_sky_cluster):
    expected_job_status = JobStatus(
        name="",
        id="1",
        cluster="new_cluster_name",
        status="",
        metadata="",
        done=False,
        state=JobState.PENDING,
    )
    mock_gcp_cluster = Mock(spec=sky.clouds.GCP)
    mock_gcp_handler = Mock()
    mock_gcp_handler.launched_resources = Mock()
    mock_gcp_handler.launched_resources.cloud = mock_gcp_cluster
    mock_sky_gcp = Mock(spec=SkyCluster)
    mock_sky_gcp.name.return_value = "new_cluster_name"
    mock_sky_gcp.get_job.return_value = expected_job_status

    mock_runpod_cluster = Mock(spec=sky.clouds.RunPod)
    mock_runpod_handler = Mock()
    mock_runpod_handler.launched_resources = Mock()
    mock_runpod_handler.launched_resources.cloud = mock_runpod_cluster
    mock_sky_runpod = Mock(spec=SkyCluster)
    mock_sky_runpod.name.return_value = "runpod_cluster"

    mock_lambda_cluster = Mock(spec=sky.clouds.Lambda)
    mock_lambda_handler = Mock()
    mock_lambda_handler.launched_resources = Mock()
    mock_lambda_handler.launched_resources.cloud = mock_lambda_cluster
    mock_sky_lambda = Mock(spec=SkyCluster)
    mock_sky_lambda.name.return_value = "lambda_cluster"
    mock_sky_cluster.side_effect = [
        mock_sky_gcp,
        mock_sky_runpod,
        mock_sky_lambda,
    ]
    mock_sky_client.status.return_value = [
        {
            "name": "new_cluster_name",
            "status": sky.ClusterStatus.UP,
            "handle": mock_gcp_handler,
        },
        {
            "name": "runpod_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_runpod_handler,
        },
        {
            "name": "lambda_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_lambda_handler,
        },
    ]
    mock_sky_client.launch.return_value = expected_job_status
    cloud = SkyCloud("gcp")
    job_status = cloud.up_cluster(
        _get_default_job("gcp"), "new_cluster_name", custom_kwarg=1
    )
    mock_sky_client.launch.assert_called_once_with(
        _get_default_job("gcp"), "new_cluster_name", custom_kwarg=1
    )
    assert job_status == expected_job_status


def test_sky_cloud_up_cluster_no_name(mock_sky_client, mock_sky_cluster):
    expected_job_status = JobStatus(
        name="",
        id="1",
        cluster="new_cluster_name",
        status="",
        metadata="",
        done=False,
        state=JobState.PENDING,
    )
    mock_gcp_cluster = Mock(spec=sky.clouds.GCP)
    mock_gcp_handler = Mock()
    mock_gcp_handler.launched_resources = Mock()
    mock_gcp_handler.launched_resources.cloud = mock_gcp_cluster
    mock_sky_gcp = Mock(spec=SkyCluster)
    mock_sky_gcp.name.return_value = "new_cluster_name"
    mock_sky_gcp.get_job.return_value = expected_job_status

    mock_runpod_cluster = Mock(spec=sky.clouds.RunPod)
    mock_runpod_handler = Mock()
    mock_runpod_handler.launched_resources = Mock()
    mock_runpod_handler.launched_resources.cloud = mock_runpod_cluster
    mock_sky_runpod = Mock(spec=SkyCluster)
    mock_sky_runpod.name.return_value = "runpod_cluster"

    mock_lambda_cluster = Mock(spec=sky.clouds.Lambda)
    mock_lambda_handler = Mock()
    mock_lambda_handler.launched_resources = Mock()
    mock_lambda_handler.launched_resources.cloud = mock_lambda_cluster
    mock_sky_lambda = Mock(spec=SkyCluster)
    mock_sky_lambda.name.return_value = "lambda_cluster"
    mock_sky_cluster.side_effect = [
        mock_sky_gcp,
        mock_sky_runpod,
        mock_sky_lambda,
    ]
    mock_sky_client.status.return_value = [
        {
            "name": "new_cluster_name",
            "status": sky.ClusterStatus.UP,
            "handle": mock_gcp_handler,
        },
        {
            "name": "runpod_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_runpod_handler,
        },
        {
            "name": "lambda_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_lambda_handler,
        },
    ]
    mock_sky_client.launch.return_value = expected_job_status
    cloud = SkyCloud("gcp")
    job_status = cloud.up_cluster(_get_default_job("gcp"), None)
    mock_sky_client.launch.assert_called_once_with(_get_default_job("gcp"), None)
    assert job_status == expected_job_status


def test_sky_cloud_list_clusters_gcp(mock_sky_client):
    cloud = SkyCloud("gcp")
    mock_gcp_cluster = Mock(spec=sky.clouds.GCP)
    mock_gcp_handler = Mock()
    mock_gcp_handler.launched_resources = Mock()
    mock_gcp_handler.launched_resources.cloud = mock_gcp_cluster

    mock_runpod_cluster = Mock(spec=sky.clouds.RunPod)
    mock_runpod_handler = Mock()
    mock_runpod_handler.launched_resources = Mock()
    mock_runpod_handler.launched_resources.cloud = mock_runpod_cluster

    mock_lambda_cluster = Mock(spec=sky.clouds.Lambda)
    mock_lambda_handler = Mock()
    mock_lambda_handler.launched_resources = Mock()
    mock_lambda_handler.launched_resources.cloud = mock_lambda_cluster
    mock_sky_client.status.return_value = [
        {
            "name": "gcp_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_gcp_handler,
        },
        {
            "name": "stop_cluster_name",
            "status": sky.ClusterStatus.STOPPED,
            "handle": mock_gcp_handler,
        },
        {
            "name": "init_cluster_name",
            "status": sky.ClusterStatus.INIT,
            "handle": mock_gcp_handler,
        },
        {
            "name": "runpod_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_runpod_handler,
        },
        {
            "name": "lambda_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_lambda_handler,
        },
    ]
    clusters = cloud.list_clusters()
    mock_sky_client.status.assert_called_once()
    assert clusters == [SkyCluster("gcp_cluster", mock_sky_client)]


def test_sky_cloud_list_clusters_runpod(mock_sky_client):
    cloud = SkyCloud("runpod")
    mock_gcp_cluster = Mock(spec=sky.clouds.GCP)
    mock_gcp_handler = Mock()
    mock_gcp_handler.launched_resources = Mock()
    mock_gcp_handler.launched_resources.cloud = mock_gcp_cluster

    mock_runpod_cluster = Mock(spec=sky.clouds.RunPod)
    mock_runpod_handler = Mock()
    mock_runpod_handler.launched_resources = Mock()
    mock_runpod_handler.launched_resources.cloud = mock_runpod_cluster

    mock_lambda_cluster = Mock(spec=sky.clouds.Lambda)
    mock_lambda_handler = Mock()
    mock_lambda_handler.launched_resources = Mock()
    mock_lambda_handler.launched_resources.cloud = mock_lambda_cluster
    mock_sky_client.status.return_value = [
        {
            "name": "gcp_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_gcp_handler,
        },
        {
            "name": "runpod_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_runpod_handler,
        },
        {
            "name": "lambda_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_lambda_handler,
        },
    ]
    clusters = cloud.list_clusters()
    mock_sky_client.status.assert_called_once()
    assert clusters == [SkyCluster("runpod_cluster", mock_sky_client)]


def test_sky_cloud_list_clusters_lambda(mock_sky_client):
    cloud = SkyCloud("lambda")
    mock_gcp_cluster = Mock(spec=sky.clouds.GCP)
    mock_gcp_handler = Mock()
    mock_gcp_handler.launched_resources = Mock()
    mock_gcp_handler.launched_resources.cloud = mock_gcp_cluster

    mock_runpod_cluster = Mock(spec=sky.clouds.RunPod)
    mock_runpod_handler = Mock()
    mock_runpod_handler.launched_resources = Mock()
    mock_runpod_handler.launched_resources.cloud = mock_runpod_cluster

    mock_lambda_cluster = Mock(spec=sky.clouds.Lambda)
    mock_lambda_handler = Mock()
    mock_lambda_handler.launched_resources = Mock()
    mock_lambda_handler.launched_resources.cloud = mock_lambda_cluster
    mock_sky_client.status.return_value = [
        {
            "name": "gcp_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_gcp_handler,
        },
        {
            "name": "runpod_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_runpod_handler,
        },
        {
            "name": "lambda_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_lambda_handler,
        },
    ]
    clusters = cloud.list_clusters()
    mock_sky_client.status.assert_called_once()
    assert clusters == [SkyCluster("lambda_cluster", mock_sky_client)]


def test_sky_cloud_list_clusters_lambda_no_cluster(mock_sky_client):
    cloud = SkyCloud("lambda")
    mock_gcp_cluster = Mock(spec=sky.clouds.GCP)
    mock_gcp_handler = Mock()
    mock_gcp_handler.launched_resources = Mock()
    mock_gcp_handler.launched_resources.cloud = mock_gcp_cluster

    mock_runpod_cluster = Mock(spec=sky.clouds.RunPod)
    mock_runpod_handler = Mock()
    mock_runpod_handler.launched_resources = Mock()
    mock_runpod_handler.launched_resources.cloud = mock_runpod_cluster

    mock_sky_client.status.return_value = [
        {
            "name": "gcp_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_gcp_handler,
        },
        {
            "name": "runpod_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_runpod_handler,
        },
    ]
    clusters = cloud.list_clusters()
    mock_sky_client.status.assert_called_once()
    assert clusters == []


def test_sky_cloud_list_clusters_lambda_multiple_cluster(mock_sky_client):
    cloud = SkyCloud("lambda")
    mock_gcp_cluster = Mock(spec=sky.clouds.GCP)
    mock_gcp_handler = Mock()
    mock_gcp_handler.launched_resources = Mock()
    mock_gcp_handler.launched_resources.cloud = mock_gcp_cluster

    mock_runpod_cluster = Mock(spec=sky.clouds.RunPod)
    mock_runpod_handler = Mock()
    mock_runpod_handler.launched_resources = Mock()
    mock_runpod_handler.launched_resources.cloud = mock_runpod_cluster

    mock_lambda_cluster = Mock(spec=sky.clouds.Lambda)
    mock_lambda_handler = Mock()
    mock_lambda_handler.launched_resources = Mock()
    mock_lambda_handler.launched_resources.cloud = mock_lambda_cluster

    mock_lambda_cluster2 = Mock(spec=sky.clouds.Lambda)
    mock_lambda_handler2 = Mock()
    mock_lambda_handler2.launched_resources = Mock()
    mock_lambda_handler2.launched_resources.cloud = mock_lambda_cluster2

    mock_sky_client.status.return_value = [
        {
            "name": "gcp_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_gcp_handler,
        },
        {
            "name": "another_one",
            "status": sky.ClusterStatus.UP,
            "handle": mock_lambda_handler2,
        },
        {
            "name": "runpod_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_runpod_handler,
        },
        {
            "name": "lambda_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_lambda_handler,
        },
    ]
    clusters = cloud.list_clusters()
    mock_sky_client.status.assert_called_once()
    assert clusters == [
        SkyCluster("another_one", mock_sky_client),
        SkyCluster("lambda_cluster", mock_sky_client),
    ]


def test_sky_cloud_list_clusters_invalid_cloud(mock_sky_client):
    cloud = SkyCloud("fake_cloud")
    mock_gcp_cluster = Mock(spec=sky.clouds.GCP)
    mock_gcp_handler = Mock()
    mock_gcp_handler.launched_resources = Mock()
    mock_gcp_handler.launched_resources.cloud = mock_gcp_cluster

    mock_runpod_cluster = Mock(spec=sky.clouds.RunPod)
    mock_runpod_handler = Mock()
    mock_runpod_handler.launched_resources = Mock()
    mock_runpod_handler.launched_resources.cloud = mock_runpod_cluster

    mock_lambda_cluster = Mock(spec=sky.clouds.Lambda)
    mock_lambda_handler = Mock()
    mock_lambda_handler.launched_resources = Mock()
    mock_lambda_handler.launched_resources.cloud = mock_lambda_cluster

    mock_lambda_cluster2 = Mock(spec=sky.clouds.Lambda)
    mock_lambda_handler2 = Mock()
    mock_lambda_handler2.launched_resources = Mock()
    mock_lambda_handler2.launched_resources.cloud = mock_lambda_cluster2

    mock_sky_client.status.return_value = [
        {
            "name": "gcp_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_gcp_handler,
        },
        {
            "name": "another_one",
            "status": sky.ClusterStatus.UP,
            "handle": mock_lambda_handler2,
        },
        {
            "name": "runpod_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_runpod_handler,
        },
        {
            "name": "lambda_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_lambda_handler,
        },
    ]
    with pytest.raises(ValueError):
        _ = cloud.list_clusters()


def test_sky_cloud_get_cluster_gcp_success(mock_sky_client):
    cloud = SkyCloud("gcp")
    mock_gcp_cluster = Mock(spec=sky.clouds.GCP)
    mock_gcp_handler = Mock()
    mock_gcp_handler.launched_resources = Mock()
    mock_gcp_handler.launched_resources.cloud = mock_gcp_cluster

    mock_runpod_cluster = Mock(spec=sky.clouds.RunPod)
    mock_runpod_handler = Mock()
    mock_runpod_handler.launched_resources = Mock()
    mock_runpod_handler.launched_resources.cloud = mock_runpod_cluster

    mock_lambda_cluster = Mock(spec=sky.clouds.Lambda)
    mock_lambda_handler = Mock()
    mock_lambda_handler.launched_resources = Mock()
    mock_lambda_handler.launched_resources.cloud = mock_lambda_cluster

    mock_sky_client.status.return_value = [
        {
            "name": "gcp_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_gcp_handler,
        },
        {
            "name": "runpod_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_runpod_handler,
        },
        {
            "name": "lambda_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_lambda_handler,
        },
    ]
    cluster = cloud.get_cluster("gcp_cluster")
    mock_sky_client.status.assert_called_once()
    assert cluster == SkyCluster("gcp_cluster", mock_sky_client)


def test_sky_cloud_get_cluster_runpod_success(mock_sky_client):
    cloud = SkyCloud("runpod")
    mock_gcp_cluster = Mock(spec=sky.clouds.GCP)
    mock_gcp_handler = Mock()
    mock_gcp_handler.launched_resources = Mock()
    mock_gcp_handler.launched_resources.cloud = mock_gcp_cluster

    mock_runpod_cluster = Mock(spec=sky.clouds.RunPod)
    mock_runpod_handler = Mock()
    mock_runpod_handler.launched_resources = Mock()
    mock_runpod_handler.launched_resources.cloud = mock_runpod_cluster

    mock_lambda_cluster = Mock(spec=sky.clouds.Lambda)
    mock_lambda_handler = Mock()
    mock_lambda_handler.launched_resources = Mock()
    mock_lambda_handler.launched_resources.cloud = mock_lambda_cluster

    mock_sky_client.status.return_value = [
        {
            "name": "gcp_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_gcp_handler,
        },
        {
            "name": "runpod_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_runpod_handler,
        },
        {
            "name": "lambda_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_lambda_handler,
        },
    ]
    cluster = cloud.get_cluster("runpod_cluster")
    mock_sky_client.status.assert_called_once()
    assert cluster == SkyCluster("runpod_cluster", mock_sky_client)


def test_sky_cloud_get_cluster_lambda_success(mock_sky_client):
    cloud = SkyCloud("lambda")
    mock_gcp_cluster = Mock(spec=sky.clouds.GCP)
    mock_gcp_handler = Mock()
    mock_gcp_handler.launched_resources = Mock()
    mock_gcp_handler.launched_resources.cloud = mock_gcp_cluster

    mock_runpod_cluster = Mock(spec=sky.clouds.RunPod)
    mock_runpod_handler = Mock()
    mock_runpod_handler.launched_resources = Mock()
    mock_runpod_handler.launched_resources.cloud = mock_runpod_cluster

    mock_lambda_cluster = Mock(spec=sky.clouds.Lambda)
    mock_lambda_handler = Mock()
    mock_lambda_handler.launched_resources = Mock()
    mock_lambda_handler.launched_resources.cloud = mock_lambda_cluster

    mock_sky_client.status.return_value = [
        {
            "name": "gcp_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_gcp_handler,
        },
        {
            "name": "runpod_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_runpod_handler,
        },
        {
            "name": "lambda_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_lambda_handler,
        },
    ]
    cluster = cloud.get_cluster("lambda_cluster")
    mock_sky_client.status.assert_called_once()
    assert cluster == SkyCluster("lambda_cluster", mock_sky_client)


def test_sky_cloud_get_cluster_aws_success(mock_sky_client):
    cloud = SkyCloud("aws")
    mock_gcp_cluster = Mock(spec=sky.clouds.GCP)
    mock_gcp_handler = Mock()
    mock_gcp_handler.launched_resources = Mock()
    mock_gcp_handler.launched_resources.cloud = mock_gcp_cluster

    mock_runpod_cluster = Mock(spec=sky.clouds.RunPod)
    mock_runpod_handler = Mock()
    mock_runpod_handler.launched_resources = Mock()
    mock_runpod_handler.launched_resources.cloud = mock_runpod_cluster

    mock_aws_cluster = Mock(spec=sky.clouds.AWS)
    mock_aws_handler = Mock()
    mock_aws_handler.launched_resources = Mock()
    mock_aws_handler.launched_resources.cloud = mock_aws_cluster

    mock_sky_client.status.return_value = [
        {
            "name": "gcp_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_gcp_handler,
        },
        {
            "name": "runpod_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_runpod_handler,
        },
        {
            "name": "aws_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_aws_handler,
        },
    ]
    cluster = cloud.get_cluster("aws_cluster")
    mock_sky_client.status.assert_called_once()
    assert cluster == SkyCluster("aws_cluster", mock_sky_client)


def test_sky_cloud_get_cluster_azure_success(mock_sky_client):
    cloud = SkyCloud("azure")
    mock_gcp_cluster = Mock(spec=sky.clouds.GCP)
    mock_gcp_handler = Mock()
    mock_gcp_handler.launched_resources = Mock()
    mock_gcp_handler.launched_resources.cloud = mock_gcp_cluster

    mock_runpod_cluster = Mock(spec=sky.clouds.RunPod)
    mock_runpod_handler = Mock()
    mock_runpod_handler.launched_resources = Mock()
    mock_runpod_handler.launched_resources.cloud = mock_runpod_cluster

    mock_azure_cluster = Mock(spec=sky.clouds.Azure)
    mock_azure_handler = Mock()
    mock_azure_handler.launched_resources = Mock()
    mock_azure_handler.launched_resources.cloud = mock_azure_cluster

    mock_sky_client.status.return_value = [
        {
            "name": "gcp_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_gcp_handler,
        },
        {
            "name": "runpod_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_runpod_handler,
        },
        {
            "name": "azure_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_azure_handler,
        },
    ]
    cluster = cloud.get_cluster("azure_cluster")
    mock_sky_client.status.assert_called_once()
    assert cluster == SkyCluster("azure_cluster", mock_sky_client)


def test_sky_cloud_get_cluster_failure_wrong_cloud(mock_sky_client):
    cloud = SkyCloud("gcp")

    mock_runpod_cluster = Mock(spec=sky.clouds.RunPod)
    mock_runpod_handler = Mock()
    mock_runpod_handler.launched_resources = Mock()
    mock_runpod_handler.launched_resources.cloud = mock_runpod_cluster

    mock_lambda_cluster = Mock(spec=sky.clouds.Lambda)
    mock_lambda_handler = Mock()
    mock_lambda_handler.launched_resources = Mock()
    mock_lambda_handler.launched_resources.cloud = mock_lambda_cluster

    mock_sky_client.status.return_value = [
        {
            "name": "runpod_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_runpod_handler,
        },
        {
            "name": "lambda_cluster",
            "status": sky.ClusterStatus.UP,
            "handle": mock_lambda_handler,
        },
    ]
    cluster = cloud.get_cluster("runpod_cluster")
    mock_sky_client.status.assert_called_once()
    assert cluster is None


def test_sky_cloud_get_cluster_failure_empty(mock_sky_client):
    cloud = SkyCloud("gcp")
    mock_sky_client.status.return_value = []
    cluster = cloud.get_cluster("gcp_cluster")
    mock_sky_client.status.assert_called_once()
    assert cluster is None


def test_runpod_cloud_builder_registered():
    assert REGISTRY.contains("runpod", RegistryType.CLOUD)


def test_gcp_cloud_builder_registered():
    assert REGISTRY.contains("gcp", RegistryType.CLOUD)


def test_lambda_cloud_builder_registered():
    assert REGISTRY.contains("lambda", RegistryType.CLOUD)


def test_aws_cloud_builder_registered():
    assert REGISTRY.contains("aws", RegistryType.CLOUD)


def test_azure_cloud_builder_registered():
    assert REGISTRY.contains("azure", RegistryType.CLOUD)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Datasets module for the Oumi (Open Universal Machine Intelligence) library.

This module provides various dataset implementations for use in the Oumi framework.
These datasets are designed for different machine learning tasks and can be used
with the models and training pipelines provided by Oumi.

For more information on the available datasets and their usage, see the
:mod:`oumi.datasets` documentation.

Each dataset is implemented as a separate class, inheriting from appropriate base
classes in the :mod:`oumi.core.datasets` module. For usage examples and more detailed
information on each dataset, please refer to their respective class documentation.

See Also:
    - :mod:`oumi.models`: Compatible models for use with these datasets.
    - :mod:`oumi.core.datasets`: Base classes for dataset implementations.

Example:
    >>> from oumi.datasets import AlpacaDataset
    >>> from torch.utils.data import DataLoader
    >>> dataset = AlpacaDataset()
    >>> train_loader = DataLoader(dataset, batch_size=32)
"""

from oumi.datasets.debug import (
    DebugClassificationDataset,
    DebugPretrainingDataset,
    DebugSftDataset,
)
from oumi.datasets.evaluation import AlpacaEvalDataset
from oumi.datasets.grpo.letter_count import LetterCountGrpoDataset
from oumi.datasets.grpo.tldr import TldrGrpoDataset
from oumi.datasets.preference_tuning.orpo_dpo_mix import OrpoDpoMix40kDataset
from oumi.datasets.pretraining.c4 import C4Dataset
from oumi.datasets.pretraining.dolma import DolmaDataset
from oumi.datasets.pretraining.falcon_refinedweb import FalconRefinedWebDataset
from oumi.datasets.pretraining.fineweb_edu import FineWebEduDataset
from oumi.datasets.pretraining.pile import PileV1Dataset
from oumi.datasets.pretraining.red_pajama_v1 import RedPajamaDataV1Dataset
from oumi.datasets.pretraining.red_pajama_v2 import RedPajamaDataV2Dataset
from oumi.datasets.pretraining.slim_pajama import SlimPajamaDataset
from oumi.datasets.pretraining.starcoder import StarCoderDataset
from oumi.datasets.pretraining.the_stack import TheStackDataset
from oumi.datasets.pretraining.tiny_stories import TinyStoriesDataset
from oumi.datasets.pretraining.tiny_textbooks import TinyTextbooksDataset
from oumi.datasets.pretraining.wikipedia import WikipediaDataset
from oumi.datasets.pretraining.wikitext import WikiTextDataset
from oumi.datasets.pretraining.youtube_commons import YouTubeCommonsDataset
from oumi.datasets.sft.alpaca import AlpacaDataset
from oumi.datasets.sft.aya import AyaDataset
from oumi.datasets.sft.chatqa import ChatqaDataset, ChatqaTatqaDataset
from oumi.datasets.sft.chatrag_bench import ChatRAGBenchDataset
from oumi.datasets.sft.coalm import CoALMDataset
from oumi.datasets.sft.dolly import ArgillaDollyDataset
from oumi.datasets.sft.huggingface import HuggingFaceDataset
from oumi.datasets.sft.magpie import ArgillaMagpieUltraDataset, MagpieProDataset
from oumi.datasets.sft.openo1_sft import OpenO1SFTDataset
from oumi.datasets.sft.prompt_response import PromptResponseDataset
from oumi.datasets.sft.sft_jsonlines import TextSftJsonLinesDataset
from oumi.datasets.sft.tulu3_sft_mixture import Tulu3MixtureDataset
from oumi.datasets.sft.ultrachat import UltrachatH4Dataset
from oumi.datasets.sft.wildchat import WildChatDataset
from oumi.datasets.vision_language.coco_captions import COCOCaptionsDataset
from oumi.datasets.vision_language.docmatix import DocmatixDataset
from oumi.datasets.vision_language.flickr30k import Flickr30kDataset
from oumi.datasets.vision_language.geometry3k import Geometry3kDataset
from oumi.datasets.vision_language.huggingface import HuggingFaceVisionDataset
from oumi.datasets.vision_language.llava_instruct_mix_vsft import (
    LlavaInstructMixVsftDataset,
)
from oumi.datasets.vision_language.lmms_lab_multimodal_open_r1 import (
    LmmsLabMultimodalOpenR1Dataset,
)
from oumi.datasets.vision_language.mnist_sft import MnistSftDataset
from oumi.datasets.vision_language.pixmo_ask_model_anything import (
    PixmoAskModelAnythingDataset,
)
from oumi.datasets.vision_language.pixmo_cap import PixmoCapDataset
from oumi.datasets.vision_language.pixmo_cap_qa import PixmoCapQADataset
from oumi.datasets.vision_language.rlaif_v import OpenbmbRlaifVDataset
from oumi.datasets.vision_language.the_cauldron import TheCauldronDataset
from oumi.datasets.vision_language.vision_dpo_jsonlines import (
    VisionDpoJsonlinesDataset,
)
from oumi.datasets.vision_language.vision_jsonlines import VLJsonlinesDataset
from oumi.datasets.vision_language.vqav2_small import Vqav2SmallDataset

__all__ = [
    "AlpacaDataset",
    "AlpacaEvalDataset",
    "ArgillaDollyDataset",
    "ArgillaMagpieUltraDataset",
    "AyaDataset",
    "C4Dataset",
    "ChatqaDataset",
    "ChatqaTatqaDataset",
    "ChatRAGBenchDataset",
    "CoALMDataset",
    "COCOCaptionsDataset",
    "DebugClassificationDataset",
    "DebugPretrainingDataset",
    "DebugSftDataset",
    "DocmatixDataset",
    "DolmaDataset",
    "FalconRefinedWebDataset",
    "FineWebEduDataset",
    "Flickr30kDataset",
    "Flickr30kDataset",
    "Geometry3kDataset",
    "HuggingFaceDataset",
    "HuggingFaceVisionDataset",
    "LetterCountGrpoDataset",
    "LlavaInstructMixVsftDataset",
    "LmmsLabMultimodalOpenR1Dataset",
    "MagpieProDataset",
    "MnistSftDataset",
    "OpenbmbRlaifVDataset",
    "OpenO1SFTDataset",
    "OrpoDpoMix40kDataset",
    "PileV1Dataset",
    "PixmoAskModelAnythingDataset",
    "PixmoCapDataset",
    "PixmoCapQADataset",
    "PromptResponseDataset",
    "RedPajamaDataV1Dataset",
    "RedPajamaDataV2Dataset",
    "SlimPajamaDataset",
    "StarCoderDataset",
    "TextSftJsonLinesDataset",
    "TheCauldronDataset",
    "TheStackDataset",
    "TinyStoriesDataset",
    "TinyTextbooksDataset",
    "TldrGrpoDataset",
    "Tulu3MixtureDataset",
    "UltrachatH4Dataset",
    "VisionDpoJsonlinesDataset",
    "VLJsonlinesDataset",
    "Vqav2SmallDataset",
    "WikipediaDataset",
    "WikiTextDataset",
    "WildChatDataset",
    "YouTubeCommonsDataset",
]



import pytest

from oumi.utils.io_utils import save_json
from oumi.utils.peft_utils import get_lora_rank


def test_get_lora_rank_fail_no_r(tmp_path):
    data = {"base_model_name_or_path": "meta-llama/Llama-3.2-3B-Instruct"}
    save_json(data, tmp_path / "adapter_config.json")

    with pytest.raises(ValueError, match="LoRA rank not found in adapter config:"):
        get_lora_rank(tmp_path)


def test_get_lora_rank_fail_r_not_int(tmp_path):
    data = {"base_model_name_or_path": "meta-llama/Llama-3.2-3B-Instruct", "r": "foo"}
    save_json(data, tmp_path / "adapter_config.json")

    with pytest.raises(ValueError, match="LoRA rank in adapter config not an int:"):
        get_lora_rank(tmp_path)


def test_get_lora_rank_successful(root_testdata_dir):
    adapter_config_path = root_testdata_dir / "adapter_config.json"
    adapter_dir = adapter_config_path.parent
    assert get_lora_rank(adapter_dir) == 64



from unittest.mock import patch

import pytest

from oumi.utils.version_utils import is_dev_build


@pytest.fixture
def mock_version():
    with patch("oumi.utils.version_utils.version") as version_mock:
        yield version_mock


def test_is_dev_build_success(mock_version):
    mock_version.return_value = "0.1.0.dev0"
    assert is_dev_build()


def test_is_dev_build_failure(mock_version):
    mock_version.return_value = "0.1.0"
    assert not is_dev_build()



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pretraining datasets module."""

from oumi.datasets.pretraining.c4 import C4Dataset
from oumi.datasets.pretraining.dolma import DolmaDataset
from oumi.datasets.pretraining.falcon_refinedweb import FalconRefinedWebDataset
from oumi.datasets.pretraining.fineweb_edu import FineWebEduDataset
from oumi.datasets.pretraining.pile import PileV1Dataset
from oumi.datasets.pretraining.red_pajama_v1 import RedPajamaDataV1Dataset
from oumi.datasets.pretraining.red_pajama_v2 import RedPajamaDataV2Dataset
from oumi.datasets.pretraining.slim_pajama import SlimPajamaDataset
from oumi.datasets.pretraining.starcoder import StarCoderDataset
from oumi.datasets.pretraining.the_stack import TheStackDataset
from oumi.datasets.pretraining.tiny_stories import TinyStoriesDataset
from oumi.datasets.pretraining.tiny_textbooks import TinyTextbooksDataset
from oumi.datasets.pretraining.wikipedia import WikipediaDataset
from oumi.datasets.pretraining.wikitext import WikiTextDataset
from oumi.datasets.pretraining.youtube_commons import YouTubeCommonsDataset

__all__ = [
    "C4Dataset",
    "DolmaDataset",
    "FalconRefinedWebDataset",
    "FineWebEduDataset",
    "PileV1Dataset",
    "RedPajamaDataV1Dataset",
    "RedPajamaDataV2Dataset",
    "SlimPajamaDataset",
    "StarCoderDataset",
    "TheStackDataset",
    "TinyStoriesDataset",
    "TinyTextbooksDataset",
    "WikipediaDataset",
    "WikiTextDataset",
    "YouTubeCommonsDataset",
]



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing_extensions import override

from oumi.core.datasets import VisionLanguageSftDataset
from oumi.core.registry import register_dataset
from oumi.core.types.conversation import (
    ContentItem,
    Conversation,
    Message,
    Role,
    Type,
)
from oumi.utils.logging import logger


@register_dataset("HuggingFaceH4/llava-instruct-mix-vsft")
class LlavaInstructMixVsftDataset(VisionLanguageSftDataset):
    """Dataset class for the `HuggingFaceH4/llava-instruct-mix-vsft` dataset."""

    default_dataset = "HuggingFaceH4/llava-instruct-mix-vsft"

    def _process_text_value(self, s: str) -> str:
        # The data contains occasional `\n` at the beginning or end
        # of text values. Let's strip them.
        return s.strip() if s else ""

    def _parse_user_messages(
        self, message_list: list[dict], images: list[dict]
    ) -> Message:
        role = Role.USER
        if len(message_list) not in (1, 2):
            raise ValueError(
                f"The `content` field for '{role}' must "
                f"contain 1 or 2 elements (question, and, optionally, image). "
                f"Actual: {len(message_list)}"
            )

        text_items: list[ContentItem] = []
        image_items: list[ContentItem] = []
        for user_message in message_list:
            message_type = user_message["type"]
            if message_type == "text":
                text_items.append(
                    ContentItem(
                        type=Type.TEXT,
                        content=self._process_text_value(user_message["text"]),
                    )
                )
            elif message_type == "image":
                image_index = int(user_message["index"])
                if not (image_index >= 0 and image_index < len(images)):
                    raise ValueError(
                        f"Image index is out-of-bounds. "
                        f"Index: {image_index} "
                        f"Image count: {len(images)}"
                    )
                image_dict = images[image_index]
                if "bytes" in image_dict and image_dict["bytes"]:
                    image_items.append(
                        ContentItem(
                            type=Type.IMAGE_BINARY,
                            binary=image_dict["bytes"],
                        )
                    )
                elif "path" in image_dict and image_dict["path"]:
                    image_items.append(
                        ContentItem(
                            type=Type.IMAGE_PATH,
                            content=image_dict["path"],
                        )
                    )
                else:
                    raise ValueError(
                        f"Image element must include 'bytes' or 'path'. "
                        f"Actual keys: {image_dict.keys()}"
                    )
            else:
                raise ValueError(
                    f"{role}'s question has unknown type: '{message_type}'"
                )

        if len(text_items) != 1:
            raise ValueError(
                f"{role}'s turn must include 1 text question. Actual: {len(text_items)}"
            )
        if len(image_items) > 1:
            raise ValueError(
                f"{role}'s turn must include max 1 image. Actual: {len(image_items)}"
            )

        # Add image messages before text messages!
        return Message(role=role, content=(image_items + text_items))

    def _parse_assistant_messages(self, message_list: list[dict]) -> Message:
        role = Role.ASSISTANT
        if len(message_list) != 1:
            raise ValueError(
                f"The `content` field for {role} must "
                f"contain exactly 1 element (response). "
                f"Actual: {len(message_list)}"
            )
        response_type = message_list[0]["type"]
        if response_type != "text":
            raise ValueError(
                f"{role}'s response is expected to be text. Actual: {response_type}"
            )

        return Message(
            role=role,
            content=self._process_text_value(message_list[0]["text"]),
        )

    @override
    def transform_conversation(self, example: dict) -> Conversation:
        """Transform a dataset example into a Conversation object."""
        example_messages = example.get("messages")

        if example_messages is None or len(example_messages) == 0:
            raise ValueError("No messages in input example.")

        images = example.get("images")
        if images is None or len(images) == 0:
            raise ValueError("No images in input example.")
        elif len(images) != 1:
            logger.warning(f"Example contains multiple images: {len(images)}")

        messages: list[Message] = []
        for message in example_messages:
            message_list = message.get("content")
            if message_list is None or len(message_list) == 0:
                raise ValueError("Missing or empty `content` field in message.")

            if message["role"] == "user":
                messages.append(self._parse_user_messages(message_list, images))
            elif message["role"] == "assistant":
                messages.append(self._parse_assistant_messages(message_list))
            else:
                raise ValueError(f"Unknown role: {message['from']}")

        return Conversation(messages=messages)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A framework used for registering and accessing objects across Oumi."""

from oumi.core.registry.registry import (
    REGISTRY,
    Registry,
    RegistryType,
    register,
    register_cloud_builder,
    register_dataset,
    register_evaluation_function,
    register_sample_analyzer,
)

__all__ = [
    "REGISTRY",
    "Registry",
    "RegistryType",
    "register",
    "register_cloud_builder",
    "register_dataset",
    "register_evaluation_function",
    "register_sample_analyzer",
]



import glob
import os
import re
from multiprocessing import Process, set_start_method
from pathlib import Path
from typing import Optional

import pytest


def _backtrack_on_path(path, n):
    """Goes up n directories in the current path."""
    output_path = path
    for _ in range(n):
        output_path = os.path.dirname(output_path)
    return output_path


def _get_oumi_path_recursively(path: Path) -> str:
    """Recursively goes up the path until it finds the oumi dir."""
    if len(path.name) == 0:
        raise FileNotFoundError("Could not find oumi dir.")
    if path.name == "oumi":
        return path.name
    return f"{_get_oumi_path_recursively(path.parent)}.{path.stem}"


def _get_all_py_paths(exclude_patterns: Optional[set[str]]) -> list[str]:
    """Recursively returns all py files in the /src/oumi/ dir of the repo."""
    path_to_current_file = os.path.realpath(__file__)
    repo_root = _backtrack_on_path(path_to_current_file, 4)
    py_pattern = str(Path(repo_root) / "src" / "oumi" / "**" / "*.py")
    all_py_files = glob.glob(py_pattern, recursive=True)
    exclude_files = []
    if exclude_patterns:
        for file in all_py_files:
            for exclude_pattern in exclude_patterns:
                if re.fullmatch(exclude_pattern, file):
                    exclude_files.append(file)
                    break
    all_py_files = [
        _get_oumi_path_recursively(Path(file))
        for file in all_py_files
        if file not in exclude_files
    ]
    assert len(all_py_files) > 0, "No py files found to parse."
    return all_py_files


all_paths = _get_all_py_paths(
    exclude_patterns={".*__init__.py", str(Path(".*") / "experimental" / ".*")},
)


def _load_module(module_path: str):
    """Loads a module from a path."""
    import importlib

    return importlib.import_module(module_path)


@pytest.mark.parametrize(
    "py_path",
    all_paths,
    ids=all_paths,
)
def test_load_py_files(py_path: str):
    # Load all python files in our src/oumi/ directory.
    # Circular dependencies will manifest as an ImportError.
    set_start_method("spawn", force=True)
    # Create a new process with a clean dependency tree.
    process = Process(target=_load_module, args=[py_path])
    process.start()
    process.join()
    assert process.exitcode == 0, (
        f"Error loading {py_path}. Exitcode: {process.exitcode}"
    )



import tempfile
from pathlib import Path
from unittest.mock import Mock

import jsonlines
import pytest
from pandas.core.api import DataFrame as DataFrame

from oumi.core.types.conversation import ContentItem, Conversation, Type
from oumi.datasets.vision_language.vision_jsonlines import (
    VLJsonlinesDataset,
)

_IMAGE_TOKEN = "<image_token>"
_IMAGE_TOKEN_ID = 32001


def create_mock_processor():
    processor = Mock()
    processor.processor_name = "Salesforce/blip2-opt-2.7b"
    processor.tokenizer = Mock()
    processor.image_processor = Mock()
    processor.chat_template = None
    processor.image_token = _IMAGE_TOKEN
    processor.image_token_id = _IMAGE_TOKEN_ID
    processor.label_ignore_index = -100
    return processor


@pytest.fixture
def mock_processor():
    return create_mock_processor()


@pytest.fixture
def sample_jsonlines_data():
    return [
        {
            "messages": [
                {
                    "role": "user",
                    "type": "compound",
                    "content": [
                        {
                            "content": "Describe this image:",
                            "type": "text",
                        },
                        {
                            "content": "path/to/image.jpg",
                            "type": "image_path",
                        },
                    ],
                },
                {
                    "role": "assistant",
                    "content": "A scenic view of the puget sound.",
                    "type": "text",
                },
            ]
        }
    ]


def test_jsonlines_init_with_data(
    sample_jsonlines_data, mock_tokenizer, mock_processor
):
    dataset = VLJsonlinesDataset(
        data=sample_jsonlines_data, tokenizer=mock_tokenizer, processor=mock_processor
    )
    assert len(dataset._data) == 1
    assert ["_messages_column"] == dataset._data.columns

    conversation = dataset.conversation(0)
    assert isinstance(conversation, Conversation)
    assert len(conversation.messages) == 2
    assert isinstance(conversation.messages[0].content, list)
    assert conversation.messages[0].content == [
        ContentItem(content="Describe this image:", type=Type.TEXT),
        ContentItem(content="path/to/image.jpg", type=Type.IMAGE_PATH),
    ]
    assert conversation.messages[1].content == "A scenic view of the puget sound."


def test_jsonlines_init_with_dataset_path(
    sample_jsonlines_data, mock_tokenizer, mock_processor
):
    with tempfile.TemporaryDirectory() as folder:
        vaild_jsonlines_filename = Path(folder) / "valid_path.jsonl"
        with jsonlines.open(vaild_jsonlines_filename, mode="w") as writer:
            writer.write_all(sample_jsonlines_data)

        dataset = VLJsonlinesDataset(
            dataset_path=vaild_jsonlines_filename,
            tokenizer=mock_tokenizer,
            processor=mock_processor,
        )
        assert len(dataset._data) == 1
        assert ["_messages_column"] == dataset._data.columns

        conversation = dataset.conversation(0)

        assert isinstance(conversation, Conversation)
        assert len(conversation.messages) == 2
        assert isinstance(conversation.messages[0].content, list)
        assert conversation.messages[0].content == [
            ContentItem(content="Describe this image:", type=Type.TEXT),
            ContentItem(content="path/to/image.jpg", type=Type.IMAGE_PATH),
        ]
        assert conversation.messages[1].content == "A scenic view of the puget sound."


def test_jsonlines_init_with_invalid_input(
    sample_jsonlines_data, mock_tokenizer, mock_processor
):
    with tempfile.TemporaryDirectory() as folder:
        valid_jsonlines_filename = Path(folder) / "valid_path.jsonl"
        with jsonlines.open(valid_jsonlines_filename, mode="w") as writer:
            writer.write_all(sample_jsonlines_data)

        with pytest.raises(ValueError, match="Dataset path or data must be provided"):
            VLJsonlinesDataset(tokenizer=mock_tokenizer, processor=mock_processor)

        with pytest.raises(
            FileNotFoundError,
            match="Provided path does not exist: 'invalid_path.jsonl'.",
        ):
            VLJsonlinesDataset(
                dataset_path="invalid_path.jsonl", tokenizer=mock_tokenizer
            )

        with pytest.raises(
            ValueError,
            match="Either dataset_path or data must be provided, but not both",
        ):
            VLJsonlinesDataset(
                dataset_path=valid_jsonlines_filename, data=[], tokenizer=mock_tokenizer
            )

        # Directory ending with .jsonl
        temp_dir_name = Path(folder) / "subdir.jsonl"
        temp_dir_name.mkdir()
        with pytest.raises(
            ValueError,
            match="Provided path is a directory, expected a file",
        ):
            VLJsonlinesDataset(dataset_path=temp_dir_name, tokenizer=mock_tokenizer)

        with pytest.raises(ValueError, match="Tokenizer must be provided"):
            VLJsonlinesDataset(
                dataset_path=valid_jsonlines_filename,
            )



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class Role(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    ROLE_UNSPECIFIED: _ClassVar[Role]
    SYSTEM: _ClassVar[Role]
    USER: _ClassVar[Role]
    ASSISTANT: _ClassVar[Role]
    TOOL: _ClassVar[Role]

ROLE_UNSPECIFIED: Role
SYSTEM: Role
USER: Role
ASSISTANT: Role
TOOL: Role

class DataBlob(_message.Message):
    __slots__ = ("mime_type", "binary_data")
    MIME_TYPE_FIELD_NUMBER: _ClassVar[int]
    BINARY_DATA_FIELD_NUMBER: _ClassVar[int]
    mime_type: str
    binary_data: bytes
    def __init__(
        self, mime_type: _Optional[str] = ..., binary_data: _Optional[bytes] = ...
    ) -> None: ...

class ContentPart(_message.Message):
    __slots__ = ("type", "content", "blob")
    class Type(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        TYPE_UNSPECIFIED: _ClassVar[ContentPart.Type]
        TEXT: _ClassVar[ContentPart.Type]
        IMAGE_PATH: _ClassVar[ContentPart.Type]
        IMAGE_URL: _ClassVar[ContentPart.Type]
        IMAGE_BINARY: _ClassVar[ContentPart.Type]

    TYPE_UNSPECIFIED: ContentPart.Type
    TEXT: ContentPart.Type
    IMAGE_PATH: ContentPart.Type
    IMAGE_URL: ContentPart.Type
    IMAGE_BINARY: ContentPart.Type
    TYPE_FIELD_NUMBER: _ClassVar[int]
    CONTENT_FIELD_NUMBER: _ClassVar[int]
    BLOB_FIELD_NUMBER: _ClassVar[int]
    type: ContentPart.Type
    content: str
    blob: DataBlob
    def __init__(
        self,
        type: _Optional[_Union[ContentPart.Type, str]] = ...,
        content: _Optional[str] = ...,
        blob: _Optional[_Union[DataBlob, _Mapping]] = ...,
    ) -> None: ...

class Message(_message.Message):
    __slots__ = ("id", "role", "parts")
    ID_FIELD_NUMBER: _ClassVar[int]
    ROLE_FIELD_NUMBER: _ClassVar[int]
    PARTS_FIELD_NUMBER: _ClassVar[int]
    id: str
    role: Role
    parts: _containers.RepeatedCompositeFieldContainer[ContentPart]
    def __init__(
        self,
        id: _Optional[str] = ...,
        role: _Optional[_Union[Role, str]] = ...,
        parts: _Optional[_Iterable[_Union[ContentPart, _Mapping]]] = ...,
    ) -> None: ...

class Conversation(_message.Message):
    __slots__ = ("conversation_id", "messages", "metadata")
    class MetadataEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    CONVERSATION_ID_FIELD_NUMBER: _ClassVar[int]
    MESSAGES_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    conversation_id: str
    messages: _containers.RepeatedCompositeFieldContainer[Message]
    metadata: _containers.ScalarMap[str, str]
    def __init__(
        self,
        conversation_id: _Optional[str] = ...,
        messages: _Optional[_Iterable[_Union[Message, _Mapping]]] = ...,
        metadata: _Optional[_Mapping[str, str]] = ...,
    ) -> None: ...
